from __future__ import annotations

import contextlib
import os
import socket
import subprocess
import sys
from functools import lru_cache
from pathlib import Path, PosixPath, PurePosixPath
from typing import Iterable, Mapping

from packaging.version import InvalidVersion, Version

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.base import Tool, ToolCache
from briefcase.integrations.subprocess import SubprocessArgsT


class XauthDatabaseCreationFailure(Exception):
    """Creating a xauth database file from the user's current X auth failed."""


class Docker(Tool):
    name = "docker"
    full_name = "Docker"

    DOCKER_VERSION_MIN = "20.10"

    WRONG_DOCKER_VERSION_ERROR = """\
Briefcase requires Docker {version_min} or higher, but you are currently running
version {docker_version}. Visit:

    {install_url}

to download and install an updated version of Docker.
"""
    UNKNOWN_DOCKER_VERSION_WARNING = """
*************************************************************************
** WARNING: Unable to determine the version of Docker                  **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you
    experience problems, this is almost certainly the cause of those
    problems.

    Please report this as a bug at:

      https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

      $ docker --version

    from the command prompt.

*************************************************************************
"""
    DOCKER_INSTALLATION_STATUS_UNKNOWN_WARNING = """
*************************************************************************
** WARNING: Unable to determine if Docker is installed                 **
*************************************************************************

    Briefcase will proceed, assuming everything is OK. If you
    experience problems, this is almost certainly the cause of those
    problems.

    Please report this as a bug at:

      https://github.com/beeware/briefcase/issues/new

    In your report, please including the output from running:

      $ docker --version

    from the command prompt.

*************************************************************************
"""
    DOCKER_NOT_INSTALLED_ERROR = """\
Briefcase requires Docker, but it is not installed (or is not on your PATH).
Visit:

    {install_url}

to download and install Docker manually.

If you have installed Docker recently and are still getting this error, you may
need to restart your terminal session.
"""

    LACKS_PERMISSION_ERROR = """\
Docker has been installed, but Briefcase is unable to invoke
Docker commands. It is possible that your user does not have
permissions to invoke Docker.

See https://docs.docker.com/engine/install/linux-postinstall/
for details on configuring access to your Docker installation.
"""
    DAEMON_NOT_RUNNING_ERROR = """\
Briefcase is unable to use Docker commands. It appears the Docker
daemon is not running.

See https://docs.docker.com/config/daemon/ for details on how to
configure your Docker daemon.
"""
    GENERIC_DOCKER_ERROR = """\
Briefcase was unable to use Docker commands. Check your Docker
installation, and try again.
"""
    BUILDX_PLUGIN_MISSING = """\
Docker is installed and available for use but the buildx plugin
is not installed. Briefcase leverages the BuildKit Docker backend
to build Docker images and the buildx plugin makes this available.

See https://docs.docker.com/go/buildx/ to install the buildx plugin.
"""

    # Platform-specific template context dictionary for Docker installation details
    DOCKER_INSTALL_URL = {
        "Windows": "https://docs.docker.com/docker-for-windows/install/",
        "Darwin": "https://docs.docker.com/docker-for-mac/install/",
        "Linux": "https://docs.docker.com/engine/install/#server",
    }

    def __init__(self, tools: ToolCache, image_tag: str | None = None):
        """A wrapper for the user-installed Docker.

        :param tools: ToolCache of available tools
        :param image_tag: An optional image used to access attributes of the Docker
            environment, such as how user permissions are managed in bind mounts. A
            lightweight image will be used if one is not specified but this image is not
            at all bound to the instance.
        """
        super().__init__(tools=tools)
        self.is_user_mapped = self._is_user_mapping_enabled(image_tag)

    @classmethod
    def verify_install(
        cls,
        tools: ToolCache,
        image_tag: str | None = None,
        **kwargs,
    ) -> Docker:
        """Verify Docker is installed and operational.

        :param tools: ToolCache of available tools
        :param image_tag: An optional image used during verification to access
            attributes of the local Docker environment. This image is not bound to the
            instance and only used during instantiation.
        """
        # short circuit since already verified and available
        if hasattr(tools, "docker"):
            return tools.docker

        cls._version_compat(tools=tools)
        cls._user_access(tools=tools)
        cls._buildx_installed(tools=tools)

        tools.docker = Docker(tools=tools, image_tag=image_tag)
        return tools.docker

    @classmethod
    def _version_compat(cls, tools: ToolCache):
        """Verify Docker version is compatible."""
        try:
            # Try to get the version of docker that is installed.
            # expected output format: Docker version 25.0.2, build 29cf629\n
            docker_version = (
                tools.subprocess.check_output(
                    ["docker", "--version"],
                    env=cls.subprocess_env(),
                )
                .split("Docker version ")[1]
                .split(",")[0]
                .split("-")[0]
            )

            # Ensure Docker version is compatible
            if Version(docker_version) < Version(cls.DOCKER_VERSION_MIN):
                raise BriefcaseCommandError(
                    cls.WRONG_DOCKER_VERSION_ERROR.format(
                        version_min=cls.DOCKER_VERSION_MIN,
                        docker_version=docker_version,
                        install_url=cls.DOCKER_INSTALL_URL[tools.host_os],
                    )
                )
        except (InvalidVersion, IndexError):
            tools.logger.warning(cls.UNKNOWN_DOCKER_VERSION_WARNING)
        except subprocess.CalledProcessError:
            tools.logger.warning(cls.DOCKER_INSTALLATION_STATUS_UNKNOWN_WARNING)
        except OSError as e:
            # Docker executable doesn't exist
            raise BriefcaseCommandError(
                cls.DOCKER_NOT_INSTALLED_ERROR.format(
                    install_url=cls.DOCKER_INSTALL_URL[tools.host_os]
                )
            ) from e

    @classmethod
    def _user_access(cls, tools: ToolCache):
        """Verify Docker is operational for user."""
        try:
            # Invoke a docker command to check if the daemon is running,
            # and the user has sufficient permissions.
            # We don't care about the output, just that it succeeds.
            tools.subprocess.check_output(
                ["docker", "info"],
                env=cls.subprocess_env(),
            )
        except subprocess.CalledProcessError as e:
            failure_output = e.output
            if "permission denied while trying to connect" in failure_output:
                raise BriefcaseCommandError(cls.LACKS_PERMISSION_ERROR) from e
            elif (
                # error message on Ubuntu
                "Is the docker daemon running?" in failure_output
                # error message on macOS
                or "connect: connection refused" in failure_output
            ):
                raise BriefcaseCommandError(cls.DAEMON_NOT_RUNNING_ERROR) from e
            else:
                raise BriefcaseCommandError(cls.GENERIC_DOCKER_ERROR) from e

    @classmethod
    def _buildx_installed(cls, tools: ToolCache):
        """Verify the buildx plugin is installed."""
        try:
            tools.subprocess.check_output(
                ["docker", "buildx", "version"],
                env=cls.subprocess_env(),
            )
        except subprocess.CalledProcessError:
            raise BriefcaseCommandError(cls.BUILDX_PLUGIN_MISSING)

    def _write_test_path(self) -> Path:
        """Host system filepath to perform write test from a container."""
        return Path.cwd() / "build/container_write_test"

    def _is_user_mapping_enabled(self, image_tag: str | None = None) -> bool:
        """Determine whether Docker is mapping users between the container and the host.

        Docker can be installed in different ways on Linux that significantly impact how
        containers interact with the host system. Of particular note is ownership of
        files and directories in bind mounts (i.e. mounts using --volume).

        Traditionally, Docker would pass through the UID/GID of the user used inside a
        container as the owner of files created within the bind mount. And since the
        default user inside containers is root, the files would be owned by root on the
        host file system; this prevents later interaction with those files by the host.
        To work around this, the Dockerfile can use a step-down user with a UID and GID
        that matches the host user running Docker.

        Other installation methods of Docker, though, are not compatible with using such
        a step-down user. This includes Docker Desktop and rootless Docker (although,
        even a traditional installation of Docker Engine can be configured similarly).
        In these modes, Docker maps the host user to the root user inside the container;
        this mapping is transparent and would require changes to the host environment to
        disable, if it can be disabled at all. This allows files created in bind mounts
        inside the container to be owned on the host file system by the user running
        Docker. Additionally, though, because the host user is mapped to root inside the
        container, any files that were created by the host user in the bind mount
        outside the container are owned by root inside the container; therefore, a step-
        down user could not interact with such bind mount files inside the container.

        To accommodate these different modes, this checks which user owns a file that is
        created inside a bind mount in the container. If the owning user of that file on
        the host file system is root, then a step-down user is necessary inside
        containers. If the owning user is the host user, root should be used.

        On macOS, Docker Desktop is the only option to use Docker and user mapping
        happens differently such that any user in the container is mapped to the host
        user. Instead of leveraging user namespaces as on Linux, this user mapping
        manifests as a consequence of bind mounts being implemented as NFS shares
        between macOS and the Linux VM that Docker Desktop runs containers in. So,
        using a step-down user on macOS is effectively inconsequential.

        On Windows WSL 2, Docker Desktop operates similarly to how it does on Linux.
        However, user namespace mapping is not possible because the Docker Desktop VM
        and the WSL distro are already running in different user namespaces...and
        therefore, Docker cannot even see the users in the distro to map them in to the
        container. So, a step-down user is always used.

        ref: https://docs.docker.com/engine/security/userns-remap/

        :param image_tag: The image:tag to use to create the container for the test; if
            one is not specified, then `alpine:latest` will be used.
        :returns: True if users are being mapped; False otherwise
        """
        host_write_test_path = self._write_test_path()
        container_write_test_path = PurePosixPath(
            f"/host_write_test/{host_write_test_path.name}"
        )

        host_write_test_path.parent.mkdir(exist_ok=True)

        try:
            host_write_test_path.unlink(missing_ok=True)
        except OSError as e:
            raise BriefcaseCommandError(
                f"""\
The file path used to determine how Docker is mapping users between the host
and Docker containers already exists and cannot be automatically deleted.

    {host_write_test_path}

Delete this file and run Briefcase again.
"""
            ) from e

        image_tag = "alpine" if image_tag is None else image_tag
        write_test_mounts = [
            (host_write_test_path.parent, container_write_test_path.parent)
        ]

        # Write the test file in the bind mount inside the container
        try:
            self.check_output(
                ["touch", container_write_test_path],
                image_tag=image_tag,
                mounts=write_test_mounts,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                "Unable to determine if Docker is mapping users"
            ) from e

        # If the file is not owned by `root`, then Docker is mapping usernames
        is_user_mapped = 0 != self.tools.os.stat(host_write_test_path).st_uid

        # Delete the file inside the container since it may be owned by root on the host
        try:
            self.check_output(
                ["rm", "-f", container_write_test_path],
                image_tag=image_tag,
                mounts=write_test_mounts,
            )
        except subprocess.CalledProcessError as e:
            raise BriefcaseCommandError(
                "Unable to clean up from determining if Docker is mapping users"
            ) from e

        return is_user_mapped

    @lru_cache(maxsize=None)
    def cache_image(self, image_tag: str):
        """Ensures an image is available and cached locally.

        While many Docker commands for an image will pull that image in-line with the
        command if it isn't already cached, this pollutes the console output with
        details about pulling the image. This can be particularly troublesome when the
        output from a command run inside a container using the image is desired.

        Note: This will not update an already cached image if a newer version is
        available in the registry.

        :param image_tag: Image name/tag to pull if not locally cached
        """
        image_id = self.tools.subprocess.check_output(
            ["docker", "images", "-q", image_tag],
            env=self.subprocess_env(),
        ).strip()

        if not image_id:
            self.tools.logger.info(
                f"Downloading Docker base image for {image_tag}...",
                prefix=self.full_name,
            )
            try:
                # disable streaming so image download progress bar is shown
                self.tools.subprocess.run(
                    ["docker", "pull", image_tag],
                    check=True,
                    stream_output=False,
                    env=self.subprocess_env(),
                )
            except subprocess.CalledProcessError as e:
                raise BriefcaseCommandError(
                    f"Unable to obtain the Docker image for {image_tag}. "
                    "Is the image name correct?"
                ) from e

    def check_output(self, args: SubprocessArgsT, image_tag: str, **kwargs) -> str:
        """Run a process inside a Docker container, capturing output.

        This ensures the image is locally cached and then runs a bare Docker invocation;
        it's really only useful for running simple commands on an image, ensuring that
        the container is destroyed afterward. In most cases, you'll want to use an app
        context, rather than this.

        :param args: The list of arguments to pass to the Docker instance.
        :param image_tag: The Docker image to run
        """
        self.cache_image(image_tag)

        # Any exceptions from running the process are *not* caught.
        # This ensures that "docker.check_output()" behaves as closely to
        # "subprocess.check_output()" as possible.
        return self.tools.subprocess.check_output(
            **self.dockerize_args(args, image_tag=image_tag, **kwargs)
        )

    @classmethod
    def subprocess_env(cls, env: dict[str, str] | None = None) -> dict[str, str]:
        """Adds environment variables to the context that Docker runs in."""
        final_env = {
            # Disable the hints/recommendations that Docker prints in the console
            "DOCKER_CLI_HINTS": "false",
        }
        if env is not None:
            final_env.update(env)
        return final_env

    def dockerize_args(
        self,
        args: SubprocessArgsT,
        image_tag: str,
        interactive: bool = False,
        mounts: Iterable[tuple[str | os.PathLike, str | os.PathLike]] | None = None,
        path_map: Mapping[str | os.PathLike, str | os.PathLike] | None = None,
        env: Mapping[str, str | os.PathLike] | None = None,
        cwd: str | os.PathLike | None = None,
        add_hosts: Iterable[tuple[str, str]] | None = None,
        **subprocess_kwargs,
    ) -> dict[str, ...]:  # pragma: no-cover-if-is-windows
        """Convert arguments and environment into a Docker-compatible form.

        Converts an argument and environment specification into a form that can be used
        as arguments to invoke Docker. This involves:

         * Ensuring the run deletes the container when it exits
         * Adding volume bind mounts for specified paths
         * Injecting environment variables in to the container's session
         * Setting the current working directory for the container's session
         * Adding additional DNS entries via /etc/hosts

        :param args: The arguments for the command to be invoked
        :param image_tag: The name of the image to start the container with
        :param interactive: Start the container with stdin attached and a tty allocated
        :param mounts: Bind mounts to add to the container
        :param path_map: Mapping of files paths from the host to the container; these
            mappings are applied to all values to be dockerized
        :param env: The environment specification for the command to be executed
        :param cwd: The working directory for the command to be executed
        :param add_hosts: DNS mappings to add to the container's network
        :param subprocess_kwargs: Keyword arguments passed through in return value
        :returns: A dictionary of keyword arguments to drive subprocess calls
        """
        docker_cmdline = ["docker", "run", "--rm"]

        # Add "-it" if in interactive mode
        if interactive:
            docker_cmdline.append("-it")

        # Add volume mounts
        # The :z suffix on volume mounts allows SELinux to modify the host mount;
        # it is ignored on non-SELinux platforms
        if mounts:
            for source, target in mounts:
                docker_cmdline.extend(
                    ("--volume", f"{os.fsdecode(source)}:{os.fsdecode(target)}:z")
                )

        # Pass environment variables in as --env arguments to Docker
        if env:
            for key, value in env.items():
                docker_cmdline.extend(
                    ("--env", f"{key}={self.dockerize_path(value, path_map)}")
                )

        # Set a cwd as the working directory for the container
        if cwd:
            docker_cmdline.extend(("--workdir", self.dockerize_path(cwd, path_map)))

        # Host mappings to add to container's /etc/hosts
        if add_hosts:
            for host_name, address in add_hosts:
                docker_cmdline.extend(("--add-host", f"{host_name}:{address}"))

        # Add the image name to create the temporary container with
        docker_cmdline.append(image_tag)

        # Finally, add the command (and its arguments) to run in the container
        docker_cmdline.extend(self.dockerize_path(arg, path_map) for arg in args)

        # Augment configuration to drive the subprocess call
        subprocess_kwargs["args"] = docker_cmdline
        subprocess_kwargs["env"] = self.subprocess_env()

        return subprocess_kwargs

    def dockerize_path(
        self,
        arg: str | os.PathLike,
        path_map: Mapping[str | os.PathLike, str | os.PathLike] | None = None,
    ) -> str:  # pragma: no-cover-if-is-windows
        """Translates host file paths into the equivalent location in the docker
        filesystem as defined by a file path mapping.

        Additionally, fsdecode() is called for the input value to ensure it's something
        that can represent a path and will always be returned as a string.

        :param arg: The argument to convert to dockerized paths
        :param path_map: A mapping of host file system paths to the docker container
            file system; e.g. $HOME/.cache/briefcase -> /briefcase, such that,
            $HOME/.cache/briefcase/tools/java becomes /briefcase/tools/java
        :returns: A string where all convertible paths have been replaced
        """
        arg = os.fsdecode(arg)

        if path_map:
            for source, target in path_map.items():
                arg = arg.replace(os.fsdecode(source), os.fsdecode(target))

        return arg

    @contextlib.contextmanager
    def x11_passthrough(
        self, subprocess_kwargs: dict[str, ...]
    ) -> dict[str, ...]:  # pragma: no-cover-if-is-windows
        """Manager to expose the host's X11 server to a container.

        This allows Docker containers to use the host's X11 server with the
        authorization afforded to the current user. The user's X11 auth is copied and
        modified in a dedicated temporary xauth file. A TCP server is set up to spoof a
        new X display and proxies X commands to the current display. The XAUTHORITY
        environment variable is set to the temporary xauth file and DISPLAY is set to
        "host.docker.internal:<display number>" so that the container sends X commands
        through a Docker-provided mapping of the host's network interface to the spoofed
        display that finally proxies those commands to the actual display.

        Full docs: https://briefcase.readthedocs.io/en/latest/how-to/internal/x11passthrough.html

        :param subprocess_kwargs: Existing keywords args from the caller
        :returns: augmented keyword args for the call to subprocess
        """
        if not (DISPLAY := self.tools.os.getenv("DISPLAY")):
            raise BriefcaseCommandError(
                "The DISPLAY environment variable must be set to run an app in Docker"
            )

        # Create a TCP proxy for a spoofed X display
        proxy_popen, proxy_display_num = self._x11_tcp_proxy(DISPLAY)

        # Define the xauth database files for the spoofed display
        xauth_file_path = self._x11_proxy_display_xauth_file_path(proxy_display_num)
        docker_xauth_file_path = PurePosixPath("/tmp") / xauth_file_path.name
        try:
            # Create the xauth database file for the spoofed display
            try:
                self._x11_write_xauth_file(DISPLAY, xauth_file_path, proxy_display_num)
            except XauthDatabaseCreationFailure:
                self.tools.logger.warning(
                    """\
An X11 authentication database could not be created for the display.

Briefcase will proceed, but if access to the display is rejected, this may be why.
"""
                )
            else:
                # Add the xauth database to the container
                subprocess_kwargs.setdefault("mounts", []).append(
                    (xauth_file_path, docker_xauth_file_path)
                )
                # Tell X clients to use the xauth database to connect to the display
                subprocess_kwargs.setdefault("env", {}).update(
                    {"XAUTHORITY": docker_xauth_file_path}
                )

            # This updates the container's /etc/hosts with a mapping for
            # `host.docker.internal` to a network address for the host such that
            # network-based services on the host can be reached within the container.
            #
            # The keyword `host.docker.internal` is already always defined for
            # containers created by Docker Desktop via the DNS configured within its
            # Linux VM. However, for Docker Engine, DNS is managed by the Linux host;
            # therefore, this mapping for `host.docker.internal` must be defined each
            # time a container is started.
            #
            # The keyword `host-gateway` is interpreted and replaced with a network
            # address for the host by the Docker server when a container starts. The
            # specific network address used here will be dependent on the Docker
            # implementation, but it will likely be either the address for `docker0` or
            # and address for the host otherwise mapped through to the container.
            subprocess_kwargs.setdefault("add_hosts", []).append(
                ("host.docker.internal", "host-gateway")
            )

            # Finally, tell X clients to use the spoofed display
            subprocess_kwargs.setdefault("env", {}).update(
                {"DISPLAY": f"host.docker.internal:{proxy_display_num}"}
            )

            yield subprocess_kwargs

        finally:
            self.tools.subprocess.cleanup("X display proxy", proxy_popen)
            xauth_file_path.unlink(missing_ok=True)

    def _x11_tcp_proxy(self, DISPLAY: str) -> tuple[subprocess.Popen, int]:
        """Starts a TCP proxy as a spoofed X display.

        The TCP server's port is chosen as 6000 + the display number as defined in the
        X11 standard. In this way, when X clients attempt to make a connection for the
        chosen display number, this TCP port will be automatically evaluated by the
        client for consideration as an X display.

        The proxy is bound to 0.0.0.0 and as such will be reachable on all network
        interfaces defined for the host. Ideally, the proxy would only need to bind to
        the network bridge defined by Docker, usually `docker0`, but Docker Desktop is
        not able to use a shared network interface bridge as Docker Engine does since
        it runs containers inside a Linux VM. Instead, it attaches the host's primary
        network interface as a device on a virtual network interface bridge created by
        the VM. Because of this, the proxy must be exposed to the network at large to
        also be exposed within the container.

        :param DISPLAY: value of DISPLAY environment variable
        :returns: the Popen process for the proxy and the display number it is using
        """
        if not self.tools.shutil.which("socat"):
            raise BriefcaseCommandError(
                "Install socat to run an app for a targeted Linux distribution"
            )

        try:
            # The DISPLAY environment variable is expected to be of the form:
            #   [hostname]:[display number].[screen number]
            display_num = int(DISPLAY.split(":")[1].split(".")[0])
        except (IndexError, ValueError, AttributeError) as e:
            raise BriefcaseCommandError(
                f"Unsupported value for environment variable DISPLAY: {str(DISPLAY)!r}"
            ) from e

        proxy_display_num = self._x11_allocate_display()
        proxy_display_tcp_port = 6000 + proxy_display_num

        # If a TCP server is already running for the X display on the host, then start
        # a proxy to that TCP server for the display. While most distros do not support
        # TCP for an X display by default, X11 forwarding over SSH is facilitated by
        # opening an SSH channel on the port for the X display; so, proxying to that
        # channel allows the container to use the X display exposed by X11 forwarding.
        if self._x11_is_display_tcp(display_num):
            proxy_process = self.tools.subprocess.Popen(
                [
                    "socat",
                    f"TCP-LISTEN:{proxy_display_tcp_port},reuseaddr,fork,bind=0.0.0.0",
                    f"TCP:localhost:{6000 + display_num}",
                ]
            )

        # Otherwise, just proxy to the UNIX socket for the display
        elif self._x11_is_display_unix(display_num):
            proxy_process = self.tools.subprocess.Popen(
                [
                    "socat",
                    f"TCP-LISTEN:{proxy_display_tcp_port},reuseaddr,fork,bind=0.0.0.0",
                    f"UNIX-CONNECT:/tmp/.X11-unix/X{display_num}",
                ]
            )

        else:
            raise BriefcaseCommandError(f"Cannot find X11 display for {DISPLAY!r}")

        return proxy_process, proxy_display_num

    def _x11_proxy_display_xauth_file_path(
        self, display_num: int
    ) -> PosixPath:  # pragma: no-cover-if-is-windows
        """Path to the xauth database for the proxy display.

        The project's build directory is used to accommodate Docker Desktop. By default,
        Docker Desktop only allows bind mounts from directories/files within the user's
        home directory. So, while it may be more appropriate to use a dedicated temp
        directory, such as /tmp, this avoids having to instruct users to add that
        directory as a source of possible bind mounts.
        """
        return PosixPath.cwd() / f"build/.briefcase.docker.xauth.{display_num}"

    def _x11_allocate_display(self) -> int:
        """Finds available X display number for use; raises otherwise.

        A universal mechanism to request and reserve an X display does not exist. On
        most systems, though, there will not be many active X displays. So, this
        approach is similar to that of OpenSSH's strategy for reserving an X display
        for X11 forwarding: simply iterate over candidate display numbers using the
        first one that is not already is use.

        This check for whether a display number is already in use is limited to
        verifying a TCP server or UNIX socket does not already exist for the display.
        While other transport protocols are possible, any modern system should be using
        at least one of these transports in tandem with anything else.
        """
        # This range of candidate display numbers is arbitrary. Typically, locally
        # configured displays start at number 0 while SSH displays start at number 10.
        # So, in the common case, this is expected to return the first candidate.
        # Although, in cases where parts of this range are already allocated, the check
        # for a UNIX socket is immediate while the TCP check must actually attempt to
        # connect to the port; in practice, this should also be immediate as anything
        # listening accepts the connection or the OS rejects it since nothing is there.
        for num in range(50, 299 + 1):
            if not (self._x11_is_display_unix(num) or self._x11_is_display_tcp(num)):
                display_num = num
                break
        else:
            raise BriefcaseCommandError("Failed to allocate X11 display")

        return display_num

    def _x11_is_display_tcp(self, display_num: int) -> bool:
        """Is a TCP server running for the X display?"""
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            sock.settimeout(3)
            try:
                # returns 0 only if the connection is successful
                return 0 == sock.connect_ex(("localhost", 6000 + display_num))
            except OSError:
                return False

    def _x11_display_unix_socket_path(
        self, display_num: int
    ) -> PosixPath:  # pragma: no-cover-if-is-windows
        """Path to the UNIX socket for the display."""
        return PosixPath(f"/tmp/.X11-unix/X{display_num}")

    def _x11_is_display_unix(
        self, display_num: int
    ) -> bool:  # pragma: no-cover-if-is-windows
        """Is a UNIX socket running for the X display?"""
        try:
            return self._x11_display_unix_socket_path(display_num).is_socket()
        except OSError:
            return False

    def _x11_write_xauth_file(
        self,
        DISPLAY: str,
        xauth_file_path: Path,
        target_display_num: int,
    ) -> None:
        """Writes a xauth file using the current user's X authorization.

        :param DISPLAY: Value of DISPLAY environment variable
        :param xauth_file_path: File path to write xauth database for target display
        :param target_display_num: Number of targeted X display
        """
        if not self.tools.shutil.which("xauth"):
            raise BriefcaseCommandError(
                "Install xauth to run an app for a targeted Linux distribution"
            )

        # Extract the 32 byte X auth cookie for the current display
        # `xauth nlist` outputs the auth C struct defined by the libxau library
        try:
            x_display_auth_cookie = (
                self.tools.subprocess.check_output(["xauth", "-i", "nlist", DISPLAY])
                .splitlines()[0]
                .split(" ")[8]  # assumed to be a MIT-MAGIC-COOKIE-1
            )
        except (subprocess.CalledProcessError, IndexError) as e:
            raise XauthDatabaseCreationFailure("Failed to retrieve xauth cookie") from e

        # Initialize X authentication file
        xauth_file_path.parent.mkdir(parents=True, exist_ok=True)
        xauth_file_path.unlink(missing_ok=True)
        xauth_file_path.touch()

        # Create a xauth database for the target display using the current display's cookie
        try:
            self.tools.subprocess.check_output(
                [
                    "xauth",
                    "-i",
                    "-f",
                    xauth_file_path,
                    "add",
                    f":{target_display_num}",
                    "MIT-MAGIC-COOKIE-1",
                    x_display_auth_cookie,
                ]
            )
        except subprocess.CalledProcessError as e:
            raise XauthDatabaseCreationFailure(
                "Failed to add a xauth entry for existing cookie"
            ) from e

        # Retrieve xauth list that was just written for the target display
        try:
            xauth_list = self.tools.subprocess.check_output(
                ["xauth", "-i", "-f", xauth_file_path, "nlist"]
            )
        except subprocess.CalledProcessError as e:
            raise XauthDatabaseCreationFailure("Failed to retrieve xauth list") from e

        # When an X auth entry is added to a xauth database, the auth is tied to a
        # hostname and a display number. Since this X auth will be used to connect to
        # the host machine's X server via a proxy exposed within the container, those
        # connections will be using a different hostname for the display. Therefore,
        # this replaces the hostname in the auth entry with the "FamilyWild" tag which
        # allows using the auth for the display regardless of the targeted hostname.
        xauth_list = "\n".join("ffff" + line[4:] for line in xauth_list.splitlines())

        # Re-write xauth for target display to support any target hostname
        try:
            self.tools.subprocess.check_output(
                ["xauth", "-i", "-f", xauth_file_path, "nmerge", "-"],
                input=xauth_list,
            )
        except subprocess.CalledProcessError as e:
            raise XauthDatabaseCreationFailure(
                "Failed to merge xauth updates for FamilyWild hostname"
            ) from e


class DockerAppContext(Tool):
    name = "docker_app_context"
    full_name = "Docker"

    def __init__(self, tools: ToolCache, app: AppConfig):
        super().__init__(tools=tools)
        self.app: AppConfig = app
        self.app_base_path: Path
        self.host_bundle_path: Path
        self.host_data_path: Path
        self.image_tag: str
        self.python_version: str

    @property
    def docker_briefcase_path(self) -> PurePosixPath:
        """The briefcase data directory used inside container."""
        return PurePosixPath("/briefcase")

    @classmethod
    def verify_install(
        cls,
        tools: ToolCache,
        app: AppConfig,
        image_tag: str,
        dockerfile_path: Path,
        app_base_path: Path,
        host_bundle_path: Path,
        host_data_path: Path,
        python_version: str,
        extra_build_args: list[str] | None = None,
        **kwargs,
    ) -> DockerAppContext:  # pragma: no-cover-if-is-windows
        """Verify that docker is available as an app-bound tool.

        Creates or updates the Docker image for the app to run commands in a context for
        the App.

        :param tools: ToolCache of available tools
        :param app: Current Appconfig
        :param image_tag: Tag to assign to Docker image
        :param dockerfile_path: Dockerfile to use to build Docker image
        :param app_base_path: Base directory path for App
        :param host_bundle_path: Base directory for where App is built
        :param host_data_path: Base directory for host's Briefcase data
        :param python_version: Version of python, e.g. 3.10
        :returns: A wrapper for a Docker app context.
        """
        # short circuit since already verified and available
        if hasattr(tools[app], "app_context"):
            return tools[app].app_context

        Docker.verify(tools=tools)

        tools[app].app_context = DockerAppContext(tools=tools, app=app)
        tools[app].app_context.prepare(
            image_tag=image_tag,
            dockerfile_path=dockerfile_path,
            app_base_path=app_base_path,
            host_bundle_path=host_bundle_path,
            host_data_path=host_data_path,
            python_version=python_version,
            extra_build_args=extra_build_args,
        )
        return tools[app].app_context

    def prepare(
        self,
        image_tag: str,
        dockerfile_path: Path,
        app_base_path: Path,
        host_bundle_path: Path,
        host_data_path: Path,
        python_version: str,
        extra_build_args: list[str] | None = None,
    ):
        """Create/update the Docker image from the app's Dockerfile."""
        self.app_base_path = app_base_path
        self.host_bundle_path = host_bundle_path
        self.host_data_path = host_data_path
        self.image_tag = image_tag
        self.python_version = python_version

        self.tools.logger.info(
            "Building Docker container image...",
            prefix=self.app.app_name,
        )
        with self.tools.input.wait_bar("Building Docker image..."):
            # Install requirements for both building *and* running the app
            # (ensure a copy of system_requires is used to avoid modification)
            system_requires = getattr(self.app, "system_requires", []).copy()
            system_requires.extend(getattr(self.app, "system_runtime_requires", []))

            with self.tools.logger.context("Docker"):
                try:
                    self.tools.subprocess.run(
                        [
                            "docker",
                            "buildx",
                            "build",
                            "--load",
                            "--progress",
                            "plain",
                            "--tag",
                            self.image_tag,
                            "--file",
                            dockerfile_path,
                            "--build-arg",
                            f"SYSTEM_REQUIRES={' '.join(system_requires)}",
                            "--build-arg",
                            f"HOST_UID={self.tools.os.getuid()}",
                            "--build-arg",
                            f"HOST_GID={self.tools.os.getgid()}",
                            Path(
                                self.app_base_path,
                                *self.app.sources[0].split("/")[:-1],
                            ),
                        ]
                        + (extra_build_args if extra_build_args is not None else []),
                        check=True,
                    )
                except subprocess.CalledProcessError as e:
                    raise BriefcaseCommandError(
                        f"Error building Docker container image for {self.app.app_name}."
                    ) from e

    @contextlib.contextmanager
    def run_app_context(self, subprocess_kwargs: dict[str, ...]) -> dict[str, ...]:
        """Manager to run a Briefcase project app in a container.

        :returns: context manager to wrap the call to Popen/run/check_output()
        """
        with self.tools.docker.x11_passthrough(subprocess_kwargs) as subprocess_kwargs:
            yield subprocess_kwargs

    def run(self, args: SubprocessArgsT, **kwargs) -> None:
        """Run a process inside a Docker container."""
        # Any exceptions from running the process are *not* caught.
        # This ensures that "docker.run()" behaves as closely to
        # "subprocess.run()" as possible.
        if kwargs.get("interactive"):
            kwargs["stream_output"] = False

        with self.tools.logger.context("Docker"):
            self.tools.subprocess.run(**self._dockerize_args(args, **kwargs))

    def check_output(self, args: SubprocessArgsT, **kwargs) -> str:
        """Capture and return output from a process inside a Docker container."""
        # Any exceptions from running the process are *not* caught.
        # This ensures that "docker.check_output()" behaves as closely to
        # "subprocess.check_output()" as possible.
        return self.tools.subprocess.check_output(
            **self._dockerize_args(args, **kwargs)
        )

    def Popen(self, args: SubprocessArgsT, **kwargs) -> subprocess.Popen:
        """Open and return a process running inside a Docker container."""
        return self.tools.subprocess.Popen(**self._dockerize_args(args, **kwargs))

    def _dockerize_args(
        self, args: SubprocessArgsT, **kwargs
    ) -> dict[str, ...]:  # pragma: no-cover-if-is-windows
        """App-specific wrapper for Docker.dockerize_args().

        Uses the app's dedicated Docker image to run the container.

        Adds app's bundle directory and the Briefcase data directory as bind mounts.

        All bind mount definitions also serve as path mappings from the host in to the
        container file system; these mappings are applied to the command arguments,
        environment variable values, and the working directory.
        """
        kwargs["image_tag"] = self.image_tag

        kwargs.setdefault("mounts", []).extend(
            [
                (self.host_bundle_path, "/app"),
                (self.host_data_path, self.docker_briefcase_path),
            ]
        )

        kwargs.setdefault("path_map", {}).update(dict(kwargs["mounts"]))

        # Convert fully-qualified paths for the host's Python to the unqualified Python
        # binary available inside the container
        kwargs["path_map"][sys.executable] = f"python{self.python_version}"

        return self.tools.docker.dockerize_args(args, **kwargs)



from briefcase.bootstraps.base import BaseGuiBootstrap


def test_base_bootstrap_fields():
    """Defined fields are implemented and implemented fields are defined."""
    for field in BaseGuiBootstrap.fields:
        assert field in BaseGuiBootstrap.__dict__

    for field in [
        attr
        for attr in BaseGuiBootstrap.__dict__
        if not attr.startswith("_")
        and attr not in {"fields", "display_name_annotation", "extra_context"}
    ]:
        assert field in BaseGuiBootstrap.fields



def test_overrides_are_used(convert_command):
    (convert_command.base_path / "src/app_name").mkdir(parents=True)
    (convert_command.base_path / "src/app_name/__main__.py").write_text(
        "", encoding="utf-8"
    )
    overrides = {
        "app_name": "app_name",
        "formal_name": "formal_name",
        "source_dir": "src/app_name",
        "test_source_dir": "test_source_dir",
        "project_name": "project_name",
        "description": "description",
        "url": "https://url.com",
        "bundle": "com.bundle",
        "author": "author",
        "author_email": "author_email",
        "license": "Other",
        "leftover": "leftover",
    }
    override_input = overrides.copy()
    out = convert_command.build_app_context(override_input)
    for k, v in overrides.items():
        if k != "leftover":
            assert out[k] == v
    assert "leftover" not in out

    assert override_input == {"leftover": "leftover"}



import pytest

from briefcase.integrations.base import ToolCache


@pytest.fixture
def mock_tools(mock_tools, android_sdk, tmp_path) -> ToolCache:
    # add Android SDK to mock tools
    mock_tools.android_sdk = android_sdk
    return mock_tools



import pytest

from briefcase.integrations.subprocess import is_process_dead


@pytest.mark.parametrize("is_pid_exists, is_dead", [(True, False), (False, True)])
def test_is_process_dead(is_pid_exists, is_dead, monkeypatch):
    """Returns inverse of whether PID exists."""
    monkeypatch.setattr("psutil.pid_exists", lambda pid: is_pid_exists)
    assert is_process_dead(100) is is_dead



import os
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.macOS.app import macOSAppOpenCommand

from ....utils import create_file


@pytest.fixture
def open_command(tmp_path, first_app_config):
    command = macOSAppOpenCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.os = MagicMock(spec_set=os)
    command.tools.subprocess = MagicMock(spec_set=Subprocess)

    return command


@pytest.mark.skipif(sys.platform != "darwin", reason="macOS specific test")
def test_open(open_command, first_app_config, tmp_path):
    """On macOS, Open starts the finder on the Content folder."""
    # Create the binary that would be in the project bundle.
    create_file(
        open_command.project_path(first_app_config) / "MacOS/First App",
        "binary",
    )

    open_command(first_app_config)

    open_command.tools.subprocess.Popen.assert_called_once_with(
        [
            "open",
            tmp_path
            / "base_path"
            / "build"
            / "first-app"
            / "macos"
            / "app"
            / "First App.app"
            / "Contents",
        ]
    )



import os
import sys

import pytest


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_dockerize_args(mock_tools, my_app, tmp_path):
    """A command to run in Docker is dockerized."""
    dockerize_args = mock_tools[my_app].app_context._dockerize_args(["hello", "world"])

    assert dockerize_args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'bundle'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "briefcase/com.example.myapp:py3.X",
            "hello",
            "world",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_dockerize_args_sys_executable(mock_tools, my_app, tmp_path):
    """A command to run in Docker using the current Python is dockerized."""
    dockerize_args = mock_tools[my_app].app_context._dockerize_args(
        [sys.executable, "-m", "pip"]
    )

    assert dockerize_args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'bundle'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "briefcase/com.example.myapp:py3.X",
            "python3.X",
            "-m",
            "pip",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_dockerize_args_mounts(mock_tools, my_app, tmp_path):
    """A command to run in Docker with mounts is dockerized."""
    dockerize_args = mock_tools[my_app].app_context._dockerize_args(
        ["hello", "world"],
        mounts=[
            ("/path/to/first", "/container/first"),
            ("/path/to/second", "/container/second"),
        ],
    )

    assert dockerize_args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            "/path/to/first:/container/first:z",
            "--volume",
            "/path/to/second:/container/second:z",
            "--volume",
            f"{tmp_path / 'bundle'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "briefcase/com.example.myapp:py3.X",
            "hello",
            "world",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_dockerize_args_mounts_path(mock_tools, my_app, tmp_path):
    """A command to run in Docker with mounts and a path inside a mount is
    dockerized."""
    dockerize_args = mock_tools[my_app].app_context._dockerize_args(
        ["hello", "world", "/path/to/second/bin"],
        mounts=[
            ("/path/to/first", "/container/first"),
            ("/path/to/second", "/container/second"),
        ],
    )

    assert dockerize_args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            "/path/to/first:/container/first:z",
            "--volume",
            "/path/to/second:/container/second:z",
            "--volume",
            f"{tmp_path / 'bundle'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "briefcase/com.example.myapp:py3.X",
            "hello",
            "world",
            "/container/second/bin",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_dockerize_args_cwd(mock_tools, my_app, tmp_path):
    """A command to run in Docker with a CWD is dockerized."""
    dockerize_args = mock_tools[my_app].app_context._dockerize_args(
        ["hello", "world"],
        cwd=tmp_path / "bundle/foobar",
    )

    assert dockerize_args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'bundle'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "--workdir",
            "/app/foobar",
            "briefcase/com.example.myapp:py3.X",
            "hello",
            "world",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_dockerize_args_arg_and_env(mock_tools, my_app, tmp_path):
    """A command to run in Docker with an extra keyword arg is dockerized."""
    dockerize_args = mock_tools[my_app].app_context._dockerize_args(
        ["hello", "world"],
        env={
            "MAGIC": "True",
            "IMPORTANCE": "super high",
        },
        check=True,
        steam_output=True,
    )

    assert dockerize_args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'bundle'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "--env",
            "MAGIC=True",
            "--env",
            "IMPORTANCE=super high",
            "briefcase/com.example.myapp:py3.X",
            "hello",
            "world",
        ],
        check=True,
        steam_output=True,
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_dockerize_args_path_arg_and_env(mock_tools, my_app, tmp_path):
    """A command to run in Docker with paths inside the bundle and Briefcase data
    directory are dockerized."""
    dockerize_args = mock_tools[my_app].app_context._dockerize_args(
        ["hello", tmp_path / "location"],
        env={
            "MAGIC": "True",
            "PATH": (
                "/somewhere/safe"
                f":{tmp_path / 'briefcase' / 'tools'}"
                f":{tmp_path / 'bundle' / 'location'}"
            ),
        },
        cwd=tmp_path / "cwd",
    )

    assert dockerize_args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{tmp_path / 'bundle'}:/app:z",
            "--volume",
            f"{tmp_path / 'briefcase'}:/briefcase:z",
            "--env",
            "MAGIC=True",
            "--env",
            "PATH=/somewhere/safe:/briefcase/tools:/app/location",
            "--workdir",
            f"{tmp_path / 'cwd'}",
            "briefcase/com.example.myapp:py3.X",
            "hello",
            os.fsdecode(tmp_path / "location"),
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )



import pytest

from briefcase.integrations.subprocess import ensure_str


@pytest.mark.parametrize(
    "indata, output",
    [
        ("a string 1\na string 2", "a string 1\na string 2"),
        (b"some bytes 1\nsome bytes 2", "some bytes 1\nsome bytes 2"),
        (1024, "1024"),
    ],
)
def test_ensure_str(indata, output):
    assert ensure_str(indata) == output



import plistlib
import subprocess
from unittest import mock

import pytest

from briefcase.console import Console, Log, LogLevel
from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.iOS.xcode import iOSXcodeBuildCommand


@pytest.fixture
def build_command(tmp_path):
    return iOSXcodeBuildCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )


@pytest.mark.parametrize("tool_debug_mode", (True, False))
def test_build_app(build_command, first_app_generated, tool_debug_mode, tmp_path):
    """An iOS App can be built."""
    build_command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    # Mock the host's CPU architecture to ensure it's reflected in the Xcode call
    build_command.tools.host_arch = "weird"

    # Enable verbose tool logging
    if tool_debug_mode:
        build_command.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    build_command.build_app(first_app_generated)

    build_command.tools.subprocess.run.assert_called_with(
        [
            "xcodebuild",
            "build",
            "-project",
            tmp_path
            / "base_path"
            / "build"
            / "first-app"
            / "ios"
            / "xcode"
            / "First App.xcodeproj",
            "-destination",
            'platform="iOS Simulator"',
            "-configuration",
            "Debug",
            "-arch",
            "weird",
            "-sdk",
            "iphonesimulator",
            "-verbose" if tool_debug_mode else "-quiet",
        ],
        check=True,
        filter_func=None if tool_debug_mode else mock.ANY,
    )

    # The app metadata references the app module
    with (tmp_path / "base_path/build/first-app/ios/xcode/Info.plist").open("rb") as f:
        plist = plistlib.load(f)
        assert plist["MainModule"] == "first_app"


def test_build_app_test_mode(build_command, first_app_generated, tmp_path):
    """An iOS App can be built in test mode."""
    build_command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    # Mock the host's CPU architecture to ensure it's reflected in the Xcode call
    build_command.tools.host_arch = "weird"

    build_command.build_app(first_app_generated, test_mode=True)

    build_command.tools.subprocess.run.assert_called_with(
        [
            "xcodebuild",
            "build",
            "-project",
            tmp_path
            / "base_path"
            / "build"
            / "first-app"
            / "ios"
            / "xcode"
            / "First App.xcodeproj",
            "-destination",
            'platform="iOS Simulator"',
            "-configuration",
            "Debug",
            "-arch",
            "weird",
            "-sdk",
            "iphonesimulator",
            "-quiet",
        ],
        check=True,
        filter_func=mock.ANY,
    )

    # The app metadata has been rewritten to reference the test module
    with (tmp_path / "base_path/build/first-app/ios/xcode/Info.plist").open("rb") as f:
        plist = plistlib.load(f)
        assert plist["MainModule"] == "tests.first_app"


def test_build_app_failed(build_command, first_app_generated, tmp_path):
    """If xcodebuild fails, an error is raised."""
    # The subprocess.run() call will raise an error
    build_command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)
    build_command.tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=["xcodebuild", "..."], returncode=1
    )

    # Mock the host's CPU architecture to ensure it's reflected in the Xcode call
    build_command.tools.host_arch = "weird"

    with pytest.raises(BriefcaseCommandError):
        build_command.build_app(first_app_generated)

    build_command.tools.subprocess.run.assert_called_with(
        [
            "xcodebuild",
            "build",
            "-project",
            tmp_path
            / "base_path"
            / "build"
            / "first-app"
            / "ios"
            / "xcode"
            / "First App.xcodeproj",
            "-destination",
            'platform="iOS Simulator"',
            "-configuration",
            "Debug",
            "-arch",
            "weird",
            "-sdk",
            "iphonesimulator",
            "-quiet",
        ],
        check=True,
        filter_func=mock.ANY,
    )



import logging

import pytest

from briefcase.console import LogLevel, RichLoggingHandler
from briefcase.exceptions import UnsupportedHostError
from briefcase.integrations.git import Git


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.git = "tool"

    tool = Git.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.git


def test_unsupported_os(mock_tools):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = "wonky"

    with pytest.raises(
        UnsupportedHostError,
        match=f"{Git.name} is not supported on wonky",
    ):
        Git.verify(mock_tools)


@pytest.mark.parametrize(
    "logging_level, handler_expected",
    [
        (LogLevel.DEEP_DEBUG, True),
        (LogLevel.DEBUG, False),
        (LogLevel.VERBOSE, False),
        (LogLevel.INFO, False),
    ],
)
def test_git_stdlib_logging(mock_tools, logging_level, handler_expected):
    """A logging handler is configured for GitPython when DEEP_DEBUG is enabled."""
    mock_tools.logger.verbosity = logging_level

    Git.verify(mock_tools)

    assert handler_expected is any(
        isinstance(h, RichLoggingHandler) for h in logging.getLogger("git").handlers
    )

    # reset handlers since they are persistent
    logging.getLogger("git").handlers.clear()



from unittest.mock import MagicMock

import pytest

import briefcase.commands.new
from briefcase.bootstraps import (
    ConsoleBootstrap,
    PygameGuiBootstrap,
    PySide6GuiBootstrap,
    TogaGuiBootstrap,
)


@pytest.fixture
def mock_builtin_bootstraps():
    return {
        "Toga": TogaGuiBootstrap,
        "Console": ConsoleBootstrap,
        "PySide6": PySide6GuiBootstrap,
        "Pygame": PygameGuiBootstrap,
    }


def test_question_sequence_toga(new_command):
    """Questions are asked, a context is constructed."""

    # Prime answers for all the questions.
    new_command.input.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
        "1",  # Toga GUI toolkit
    ]

    context = new_command.build_context(
        project_overrides={},
    )

    assert context == dict(
        app_name="myapplication",
        author="Grace Hopper",
        author_email="grace@navy.mil",
        bundle="org.beeware",
        class_name="MyApplication",
        description="Cool stuff",
        formal_name="My Application",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myapplication",
        source_dir="src/myapplication",
        test_source_dir="tests",
        project_name="My Project",
        url="https://navy.mil/myapplication",
        app_source='''\
import toga
from toga.style import Pack
from toga.style.pack import COLUMN, ROW


class {{ cookiecutter.class_name }}(toga.App):
    def startup(self):
        """Construct and show the Toga application.

        Usually, you would add your application to a main content box.
        We then create a main window (with a name matching the app), and
        show the main window.
        """
        main_box = toga.Box()

        self.main_window = toga.MainWindow(title=self.formal_name)
        self.main_window.content = main_box
        self.main_window.show()


def main():
    return {{ cookiecutter.class_name }}()
''',
        app_start_source="""\
from {{ cookiecutter.module_name }}.app import main

if __name__ == "__main__":
    main().main_loop()
""",
        pyproject_table_briefcase_app_extra_content="""
requires = [
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        pyproject_table_macOS="""\
universal_build = true
requires = [
    "toga-cocoa~=0.4.5",
    "std-nslog~=1.0.0",
]
""",
        pyproject_table_linux="""\
requires = [
    "toga-gtk~=0.4.5",
]
""",
        pyproject_table_linux_system_debian="""\
system_requires = [
    # Needed to compile pycairo wheel
    "libcairo2-dev",
    # Needed to compile PyGObject wheel
    "libgirepository1.0-dev",
]

system_runtime_requires = [
    # Needed to provide GTK and its GI bindings
    "gir1.2-gtk-3.0",
    "libgirepository-1.0-1",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3-module",
    # Needed to provide WebKit2 at runtime
    # Note: Debian 11 and Ubuntu 20.04 require gir1.2-webkit2-4.0 instead
    # "gir1.2-webkit2-4.1",
]
""",
        pyproject_table_linux_system_rhel="""\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo-gobject-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
]

system_runtime_requires = [
    # Needed to support Python bindings to GTK
    "gobject-introspection",
    # Needed to provide GTK
    "gtk3",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3",
    # Needed to provide WebKit2 at runtime
    # "webkit2gtk3",
]
""",
        pyproject_table_linux_system_suse="""\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
]

system_runtime_requires = [
    # Needed to provide GTK
    "gtk3",
    # Needed to support Python bindings to GTK
    "gobject-introspection", "typelib(Gtk) = 3.0",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3-module",
    # Needed to provide WebKit2 at runtime
    # "libwebkit2gtk3", "typelib(WebKit2)",
]
""",
        pyproject_table_linux_system_arch="""\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo",
    # Needed to compile PyGObject wheel
    "gobject-introspection",
    # Runtime dependencies that need to exist so that the
    # Arch package passes final validation.
    # Needed to provide GTK
    "gtk3",
    # Dependencies that GTK looks for at runtime
    "libcanberra",
    # Needed to provide WebKit2
    # "webkit2gtk",
]

system_runtime_requires = [
    # Needed to provide GTK
    "gtk3",
    # Needed to provide PyGObject bindings
    "gobject-introspection-runtime",
    # Dependencies that GTK looks for at runtime
    "libcanberra",
    # Needed to provide WebKit2 at runtime
    # "webkit2gtk",
]
""",
        pyproject_table_linux_appimage="""\
manylinux = "manylinux_2_28"

system_requires = [
    # Needed to compile pycairo wheel
    "cairo-gobject-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
    # Needed to provide GTK
    "gtk3-devel",
    # Dependencies that GTK looks for at runtime, that need to be
    # in the build environment to be picked up by linuxdeploy
    "libcanberra-gtk3",
    "PackageKit-gtk3-module",
    "gvfs-client",
]

linuxdeploy_plugins = [
    "DEPLOY_GTK_VERSION=3 gtk",
]
""",
        pyproject_table_linux_flatpak="""\
flatpak_runtime = "org.gnome.Platform"
flatpak_runtime_version = "45"
flatpak_sdk = "org.gnome.Sdk"
""",
        pyproject_table_windows="""\
requires = [
    "toga-winforms~=0.4.5",
]
""",
        pyproject_table_iOS="""\
requires = [
    "toga-iOS~=0.4.5",
    "std-nslog~=1.0.0",
]
""",
        pyproject_table_android="""\
requires = [
    "toga-android~=0.4.5",
]

base_theme = "Theme.MaterialComponents.Light.DarkActionBar"

build_gradle_dependencies = [
    "com.google.android.material:material:1.11.0",
    # Needed for DetailedList
    # "androidx.swiperefreshlayout:swiperefreshlayout:1.1.0",
    # Needed for MapView
    # "org.osmdroid:osmdroid-android:6.1.0",
]
""",
        pyproject_table_web="""\
requires = [
    "toga-web~=0.4.5",
]
style_framework = "Shoelace v2.3"
""",
    )


def test_question_sequence_console(new_command):
    """A console app can be constructed."""

    # Prime answers for all the questions.
    new_command.input.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
        "4",  # Console app
    ]

    context = new_command.build_context(
        project_overrides={},
    )

    assert context == dict(
        app_name="myapplication",
        author="Grace Hopper",
        author_email="grace@navy.mil",
        bundle="org.beeware",
        class_name="MyApplication",
        description="Cool stuff",
        formal_name="My Application",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myapplication",
        source_dir="src/myapplication",
        test_source_dir="tests",
        project_name="My Project",
        url="https://navy.mil/myapplication",
        app_source="""\

def main():
    # Your app logic goes here
    print("Hello, World.")
""",
        app_start_source="""\
from {{ cookiecutter.module_name }}.app import main

if __name__ == "__main__":
    main()
""",
        pyproject_table_briefcase_app_extra_content="""
console_app = true
requires = [
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        pyproject_table_macOS="""\
universal_build = true
requires = [
]
""",
        pyproject_table_linux="""\
requires = [
]
""",
        pyproject_table_linux_system_debian="""\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        pyproject_table_linux_system_rhel="""\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        pyproject_table_linux_system_suse="""\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        pyproject_table_linux_system_arch="""\
system_requires = [
    # Add any system packages needed at build the app here
]

system_runtime_requires = [
    # Add any system packages needed at runtime here
]
""",
        pyproject_table_linux_flatpak="""\
flatpak_runtime = "org.freedesktop.Platform"
flatpak_runtime_version = "23.08"
flatpak_sdk = "org.freedesktop.Sdk"
""",
        pyproject_table_windows="""\
requires = [
]
""",
        pyproject_table_iOS="""\
supported = false
""",
        pyproject_table_android="""\
supported = false
""",
        pyproject_table_web="""\
supported = false
""",
    )


def test_question_sequence_pyside6(new_command):
    """Questions are asked, a context is constructed."""

    # Prime answers for all the questions.
    new_command.input.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
        "2",  # PySide6 GUI toolkit
    ]

    context = new_command.build_context(
        project_overrides={},
    )

    assert context == dict(
        app_name="myapplication",
        author="Grace Hopper",
        author_email="grace@navy.mil",
        bundle="org.beeware",
        class_name="MyApplication",
        description="Cool stuff",
        formal_name="My Application",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myapplication",
        source_dir="src/myapplication",
        test_source_dir="tests",
        project_name="My Project",
        url="https://navy.mil/myapplication",
        app_source="""\
import importlib.metadata
import sys

from PySide6 import QtWidgets


class {{ cookiecutter.class_name }}(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.init_ui()

    def init_ui(self):
        self.setWindowTitle("{{ cookiecutter.app_name }}")
        self.show()


def main():
    # Linux desktop environments use an app's .desktop file to integrate the app
    # in to their application menus. The .desktop file of this app will include
    # the StartupWMClass key, set to app's formal name. This helps associate the
    # app's windows to its menu item.
    #
    # For association to work, any windows of the app must have WMCLASS property
    # set to match the value set in app's desktop file. For PySide6, this is set
    # with setApplicationName().

    # Find the name of the module that was used to start the app
    app_module = sys.modules["__main__"].__package__
    # Retrieve the app's metadata
    metadata = importlib.metadata.metadata(app_module)

    QtWidgets.QApplication.setApplicationName(metadata["Formal-Name"])

    app = QtWidgets.QApplication(sys.argv)
    main_window = {{ cookiecutter.class_name }}()
    sys.exit(app.exec())
""",
        pyproject_table_briefcase_app_extra_content="""
requires = [
    "PySide6-Essentials~=6.5",
    # "PySide6-Addons~=6.5",
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        pyproject_table_macOS="""\
universal_build = true
requires = [
    "std-nslog~=1.0.0",
]
""",
        pyproject_table_linux="""\
requires = [
]
""",
        pyproject_table_linux_system_debian="""\
system_requires = [
]

system_runtime_requires = [
    # Derived from https://doc.qt.io/qt-6/linux-requirements.html
    "libxext6",
    "libxrender1",
    "libx11-xcb1",
    "libxkbcommon-x11-0",
    "libxcb-image0",
    "libxcb-cursor0",
    "libxcb-shape0",
    "libxcb-randr0",
    "libxcb-xfixes0",
    "libxcb-sync1",
    "libxcb-icccm4",
    "libxcb-keysyms1",
    "libfontconfig1",
    "libsm6",
    "libice6",
    "libglib2.0-0",
    "libgl1",
    "libegl1",
    "libdbus-1-3",
]
""",
        pyproject_table_linux_system_rhel="""\
system_requires = [
]

system_runtime_requires = [
    "qt6-qtbase-gui",
]
""",
        pyproject_table_linux_system_suse="""\
system_requires = [
]

system_runtime_requires = [
    "libgthread-2_0-0",
    "libQt6Gui6",
]
""",
        pyproject_table_linux_system_arch="""\
system_requires = [
]

system_runtime_requires = [
    "qt6-base",
]
""",
        pyproject_table_linux_appimage="""\
manylinux = "manylinux_2_28"

system_requires = [
# ?? FIXME
]

linuxdeploy_plugins = [
]
""",
        pyproject_table_linux_flatpak="""\
flatpak_runtime = "org.kde.Platform"
flatpak_runtime_version = "6.6"
flatpak_sdk = "org.kde.Sdk"
""",
        pyproject_table_windows="""\
requires = [
]
""",
        pyproject_table_iOS="""\
supported = false
""",
        pyproject_table_android="""\
supported = false
""",
        pyproject_table_web="""\
supported = false
""",
    )


def test_question_sequence_pygame(new_command):
    """Questions are asked, a context is constructed."""

    # Prime answers for all the questions.
    new_command.input.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
        "3",  # Pygame GUI toolkit
    ]

    context = new_command.build_context(
        project_overrides={},
    )

    assert context == dict(
        app_name="myapplication",
        author="Grace Hopper",
        author_email="grace@navy.mil",
        bundle="org.beeware",
        class_name="MyApplication",
        description="Cool stuff",
        formal_name="My Application",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myapplication",
        source_dir="src/myapplication",
        test_source_dir="tests",
        project_name="My Project",
        url="https://navy.mil/myapplication",
        app_source="""\
import importlib.metadata
import os
import sys
from pathlib import Path

import pygame


SCREEN_WIDTH, SCREEN_HEIGHT = 800, 600
WHITE = (255, 255, 255)


def main():
    # Linux desktop environments use an app's .desktop file to integrate the app
    # in to their application menus. The .desktop file of this app will include
    # the StartupWMClass key, set to app's formal name. This helps associate the
    # app's windows to its menu item.
    #
    # For association to work, any windows of the app must have WMCLASS property
    # set to match the value set in app's desktop file. For PyGame, this is set
    # using the SDL_VIDEO_X11_WMCLASS environment variable.

    # Find the name of the module that was used to start the app
    app_module = sys.modules["__main__"].__package__
    # Retrieve the app's metadata
    metadata = importlib.metadata.metadata(app_module)

    os.environ["SDL_VIDEO_X11_WMCLASS"] = metadata["Formal-Name"]

    pygame.init()
    pygame.display.set_caption(metadata["Formal-Name"])
    screen = pygame.display.set_mode((SCREEN_WIDTH, SCREEN_HEIGHT))

    running = True
    while running:
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                running = False
                break

        screen.fill(WHITE)
        pygame.display.flip()

    pygame.quit()
""",
        pyproject_table_briefcase_app_extra_content="""
requires = [
    "pygame~=2.2",
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        pyproject_table_macOS="""\
universal_build = true
requires = [
    "std-nslog~=1.0.0",
]
""",
        pyproject_table_linux="""\
requires = [
]
""",
        pyproject_table_linux_system_debian="""\
system_requires = [
]

system_runtime_requires = [
]
""",
        pyproject_table_linux_system_rhel="""\
system_requires = [
]

system_runtime_requires = [
]
""",
        pyproject_table_linux_system_suse="""\
system_requires = [
]

system_runtime_requires = [
]
""",
        pyproject_table_linux_system_arch="""\
system_requires = [
]

system_runtime_requires = [
]
""",
        pyproject_table_linux_appimage="""\
manylinux = "manylinux_2_28"

system_requires = [
]

linuxdeploy_plugins = [
]
""",
        pyproject_table_linux_flatpak="""\
flatpak_runtime = "org.freedesktop.Platform"
flatpak_runtime_version = "23.08"
flatpak_sdk = "org.freedesktop.Sdk"
""",
        pyproject_table_windows="""\
requires = [
]
""",
        pyproject_table_iOS="""\
supported = false
""",
        pyproject_table_android="""\
supported = false
""",
        pyproject_table_web="""\
supported = false
""",
    )


def test_question_sequence_none(new_command):
    """Questions are asked, a context is constructed."""

    # Prime answers for all the questions.
    new_command.input.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
        "5",  # None
    ]

    context = new_command.build_context(
        project_overrides={},
    )

    assert context == dict(
        app_name="myapplication",
        author="Grace Hopper",
        author_email="grace@navy.mil",
        bundle="org.beeware",
        class_name="MyApplication",
        description="Cool stuff",
        formal_name="My Application",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myapplication",
        source_dir="src/myapplication",
        test_source_dir="tests",
        project_name="My Project",
        url="https://navy.mil/myapplication",
    )


def test_question_sequence_with_overrides(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """Questions are asked, a context is constructed."""

    # Prime answers for none of the questions.
    new_command.input.values = []

    class GuiBootstrap:
        fields = []

        def __init__(self, context):
            pass

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value=dict(
                **mock_builtin_bootstraps,
                **{"Custom GUI": GuiBootstrap},
            ),
        ),
    )

    context = new_command.build_context(
        project_overrides=dict(
            formal_name="My Override App",
            app_name="myoverrideapp",
            bundle="net.example",
            project_name="My Override Project",
            description="My override description",
            author="override, author",
            author_email="author@override.tld",
            url="https://override.example.com",
            license="MIT license",
            bootstrap="Custom GUI",
        ),
    )

    assert context == dict(
        app_name="myoverrideapp",
        author="override, author",
        author_email="author@override.tld",
        bundle="net.example",
        class_name="MyOverrideApp",
        description="My override description",
        formal_name="My Override App",
        license="MIT license",
        module_name="myoverrideapp",
        source_dir="src/myoverrideapp",
        test_source_dir="tests",
        project_name="My Override Project",
        url="https://override.example.com",
    )


def test_question_sequence_with_bad_license_override(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """A bad override for license uses user input instead."""

    # Prime answers for all the questions.
    new_command.input.values = [
        "4",  # license
    ]

    class GuiBootstrap:
        fields = []

        def __init__(self, context):
            pass

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value=dict(
                **mock_builtin_bootstraps,
                **{"Custom GUI": GuiBootstrap},
            ),
        ),
    )

    context = new_command.build_context(
        project_overrides=dict(
            formal_name="My Override App",
            app_name="myoverrideapp",
            bundle="net.example",
            project_name="My Override Project",
            description="My override description",
            author="override, author",
            author_email="author@override.tld",
            url="https://override.example.com",
            license="BAD i don't exist license",
            bootstrap="Custom GUI",
        ),
    )

    assert context == dict(
        app_name="myoverrideapp",
        author="override, author",
        author_email="author@override.tld",
        bundle="net.example",
        class_name="MyOverrideApp",
        description="My override description",
        formal_name="My Override App",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myoverrideapp",
        source_dir="src/myoverrideapp",
        test_source_dir="tests",
        project_name="My Override Project",
        url="https://override.example.com",
    )


def test_question_sequence_with_bad_bootstrap_override(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """A bad override for the bootstrap uses user input instead."""

    # Prime answers for none of the questions.
    new_command.input.values = [
        "6",  # None
    ]

    class GuiBootstrap:
        # if this custom bootstrap is chosen, the lack of
        # requires() would cause an error
        fields = ["requires"]

        def __init__(self, context):
            pass

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value=dict(
                **mock_builtin_bootstraps,
                **{"Custom GUI": GuiBootstrap},
            ),
        ),
    )

    context = new_command.build_context(
        project_overrides=dict(
            formal_name="My Override App",
            app_name="myoverrideapp",
            bundle="net.example",
            project_name="My Override Project",
            description="My override description",
            author="override, author",
            author_email="author@override.tld",
            url="https://override.example.com",
            license="MIT license",
            bootstrap="BAD i don't exist GUI",
        ),
    )

    assert context == dict(
        app_name="myoverrideapp",
        author="override, author",
        author_email="author@override.tld",
        bundle="net.example",
        class_name="MyOverrideApp",
        description="My override description",
        formal_name="My Override App",
        license="MIT license",
        module_name="myoverrideapp",
        source_dir="src/myoverrideapp",
        test_source_dir="tests",
        project_name="My Override Project",
        url="https://override.example.com",
    )


def test_question_sequence_with_no_user_input(new_command):
    """If no user input is provided, all user inputs are taken as default."""

    new_command.input.enabled = False

    context = new_command.build_context(project_overrides={})

    assert context == dict(
        app_name="helloworld",
        author="Jane Developer",
        author_email="jane@example.com",
        bundle="com.example",
        class_name="HelloWorld",
        description="My first application",
        formal_name="Hello World",
        license="BSD license",
        module_name="helloworld",
        source_dir="src/helloworld",
        test_source_dir="tests",
        project_name="Hello World",
        url="https://example.com/helloworld",
        app_source='''\
import toga
from toga.style import Pack
from toga.style.pack import COLUMN, ROW


class {{ cookiecutter.class_name }}(toga.App):
    def startup(self):
        """Construct and show the Toga application.

        Usually, you would add your application to a main content box.
        We then create a main window (with a name matching the app), and
        show the main window.
        """
        main_box = toga.Box()

        self.main_window = toga.MainWindow(title=self.formal_name)
        self.main_window.content = main_box
        self.main_window.show()


def main():
    return {{ cookiecutter.class_name }}()
''',
        app_start_source="""\
from {{ cookiecutter.module_name }}.app import main

if __name__ == "__main__":
    main().main_loop()
""",
        pyproject_table_briefcase_app_extra_content="""
requires = [
]
test_requires = [
{% if cookiecutter.test_framework == "pytest" %}
    "pytest",
{% endif %}
]
""",
        pyproject_table_macOS="""\
universal_build = true
requires = [
    "toga-cocoa~=0.4.5",
    "std-nslog~=1.0.0",
]
""",
        pyproject_table_linux="""\
requires = [
    "toga-gtk~=0.4.5",
]
""",
        pyproject_table_linux_system_debian="""\
system_requires = [
    # Needed to compile pycairo wheel
    "libcairo2-dev",
    # Needed to compile PyGObject wheel
    "libgirepository1.0-dev",
]

system_runtime_requires = [
    # Needed to provide GTK and its GI bindings
    "gir1.2-gtk-3.0",
    "libgirepository-1.0-1",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3-module",
    # Needed to provide WebKit2 at runtime
    # Note: Debian 11 and Ubuntu 20.04 require gir1.2-webkit2-4.0 instead
    # "gir1.2-webkit2-4.1",
]
""",
        pyproject_table_linux_system_rhel="""\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo-gobject-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
]

system_runtime_requires = [
    # Needed to support Python bindings to GTK
    "gobject-introspection",
    # Needed to provide GTK
    "gtk3",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3",
    # Needed to provide WebKit2 at runtime
    # "webkit2gtk3",
]
""",
        pyproject_table_linux_system_suse="""\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
]

system_runtime_requires = [
    # Needed to provide GTK
    "gtk3",
    # Needed to support Python bindings to GTK
    "gobject-introspection", "typelib(Gtk) = 3.0",
    # Dependencies that GTK looks for at runtime
    "libcanberra-gtk3-module",
    # Needed to provide WebKit2 at runtime
    # "libwebkit2gtk3", "typelib(WebKit2)",
]
""",
        pyproject_table_linux_system_arch="""\
system_requires = [
    # Needed to compile pycairo wheel
    "cairo",
    # Needed to compile PyGObject wheel
    "gobject-introspection",
    # Runtime dependencies that need to exist so that the
    # Arch package passes final validation.
    # Needed to provide GTK
    "gtk3",
    # Dependencies that GTK looks for at runtime
    "libcanberra",
    # Needed to provide WebKit2
    # "webkit2gtk",
]

system_runtime_requires = [
    # Needed to provide GTK
    "gtk3",
    # Needed to provide PyGObject bindings
    "gobject-introspection-runtime",
    # Dependencies that GTK looks for at runtime
    "libcanberra",
    # Needed to provide WebKit2 at runtime
    # "webkit2gtk",
]
""",
        pyproject_table_linux_appimage="""\
manylinux = "manylinux_2_28"

system_requires = [
    # Needed to compile pycairo wheel
    "cairo-gobject-devel",
    # Needed to compile PyGObject wheel
    "gobject-introspection-devel",
    # Needed to provide GTK
    "gtk3-devel",
    # Dependencies that GTK looks for at runtime, that need to be
    # in the build environment to be picked up by linuxdeploy
    "libcanberra-gtk3",
    "PackageKit-gtk3-module",
    "gvfs-client",
]

linuxdeploy_plugins = [
    "DEPLOY_GTK_VERSION=3 gtk",
]
""",
        pyproject_table_linux_flatpak="""\
flatpak_runtime = "org.gnome.Platform"
flatpak_runtime_version = "45"
flatpak_sdk = "org.gnome.Sdk"
""",
        pyproject_table_windows="""\
requires = [
    "toga-winforms~=0.4.5",
]
""",
        pyproject_table_iOS="""\
requires = [
    "toga-iOS~=0.4.5",
    "std-nslog~=1.0.0",
]
""",
        pyproject_table_android="""\
requires = [
    "toga-android~=0.4.5",
]

base_theme = "Theme.MaterialComponents.Light.DarkActionBar"

build_gradle_dependencies = [
    "com.google.android.material:material:1.11.0",
    # Needed for DetailedList
    # "androidx.swiperefreshlayout:swiperefreshlayout:1.1.0",
    # Needed for MapView
    # "org.osmdroid:osmdroid-android:6.1.0",
]
""",
        pyproject_table_web="""\
requires = [
    "toga-web~=0.4.5",
]
style_framework = "Shoelace v2.3"
""",
    )


def test_question_sequence_custom_bootstrap(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """Questions are asked, a context is constructed."""

    class GuiBootstrap:
        fields = ["requires", "platform"]

        def __init__(self, context):
            pass

        def extra_context(self):
            return {"custom_context": "value"}

        def requires(self):
            return "toga"

        def platform(self):
            return "bsd"

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value=dict(
                **mock_builtin_bootstraps,
                **{"Custom GUI": GuiBootstrap},
            ),
        ),
    )

    # Prime answers for all the questions.
    new_command.input.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
        "5",  # Custom GUI bootstrap
    ]

    context = new_command.build_context(project_overrides={})

    assert context == dict(
        app_name="myapplication",
        author="Grace Hopper",
        author_email="grace@navy.mil",
        bundle="org.beeware",
        class_name="MyApplication",
        description="Cool stuff",
        formal_name="My Application",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myapplication",
        source_dir="src/myapplication",
        test_source_dir="tests",
        project_name="My Project",
        url="https://navy.mil/myapplication",
        custom_context="value",
        requires="toga",
        platform="bsd",
    )


def test_question_sequence_custom_bootstrap_without_additional_context(
    new_command,
    mock_builtin_bootstraps,
    monkeypatch,
):
    """Questions are asked, a context is constructed."""

    class GuiBootstrap:
        fields = ["requires", "platform"]

        def __init__(self, context):
            pass

        def requires(self):
            return "toga"

        def platform(self):
            return "bsd"

    monkeypatch.setattr(
        briefcase.commands.new,
        "get_gui_bootstraps",
        MagicMock(
            return_value=dict(
                **mock_builtin_bootstraps,
                **{"Custom GUI": GuiBootstrap},
            ),
        ),
    )

    # Prime answers for all the questions.
    new_command.input.values = [
        "My Application",  # formal name
        "",  # app name - accept the default
        "org.beeware",  # bundle ID
        "My Project",  # project name
        "Cool stuff",  # description
        "Grace Hopper",  # author
        "grace@navy.mil",  # author email
        "https://navy.mil/myapplication",  # URL
        "4",  # license
        "5",  # Custom GUI bootstrap
    ]

    context = new_command.build_context(project_overrides={})

    assert context == dict(
        app_name="myapplication",
        author="Grace Hopper",
        author_email="grace@navy.mil",
        bundle="org.beeware",
        class_name="MyApplication",
        description="Cool stuff",
        formal_name="My Application",
        license="GNU General Public License v2 (GPLv2)",
        module_name="myapplication",
        source_dir="src/myapplication",
        test_source_dir="tests",
        project_name="My Project",
        url="https://navy.mil/myapplication",
        requires="toga",
        platform="bsd",
    )



import pytest

from briefcase.config import AppConfig
from briefcase.exceptions import BriefcaseConfigError


def test_minimal_AppConfig():
    """A simple config can be defined."""
    config = AppConfig(
        app_name="myapp",
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        sources=["src/myapp", "somewhere/else/interesting", "local_app"],
        license={"file": "LICENSE"},
    )

    # The basic properties have been set.
    assert config.app_name == "myapp"
    assert config.version == "1.2.3"
    assert config.bundle == "org.beeware"
    assert config.description == "A simple app"
    assert config.requires is None

    # Derived properties have been set.
    assert config.bundle_name == "myapp"
    assert config.bundle_identifier == "org.beeware.myapp"
    assert config.formal_name == "myapp"
    assert config.class_name == "myapp"
    assert config.document_types == {}

    # There is no icon of any kind
    assert config.icon is None

    # The PYTHONPATH is derived correctly
    assert config.PYTHONPATH(False) == ["src", "somewhere/else", ""]
    # The test mode PYTHONPATH is the same
    assert config.PYTHONPATH(True) == ["src", "somewhere/else", ""]

    # The object has a meaningful REPL
    assert repr(config) == "<org.beeware.myapp v1.2.3 AppConfig>"


def test_extra_attrs():
    """A config can contain attributes in addition to those required."""
    config = AppConfig(
        app_name="myapp",
        formal_name="My App!",
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        long_description="A longer description\nof the app",
        license={"file": "LICENSE"},
        template="/path/to/template",
        sources=["src/myapp"],
        requires=["first", "second", "third"],
        document_type={
            "document": {
                "extension": "doc",
                "description": "A document",
            }
        },
        first="value 1",
        second=42,
    )

    # The basic properties have been set.
    assert config.app_name == "myapp"
    assert config.version == "1.2.3"
    assert config.bundle == "org.beeware"
    assert config.description == "A simple app"
    assert config.long_description == "A longer description\nof the app"
    assert config.template == "/path/to/template"
    assert config.requires == ["first", "second", "third"]

    # Properties that are derived by default have been set explicitly
    assert config.formal_name == "My App!"
    assert config.class_name == "MyApp"
    assert config.document_types == {
        "document": {
            "extension": "doc",
            "description": "A document",
        }
    }

    # Explicit additional properties have been set
    assert config.first == "value 1"
    assert config.second == 42

    # An attribute that wasn't provided raises an error
    with pytest.raises(AttributeError):
        _ = config.unknown


@pytest.mark.parametrize(
    "name",
    [
        "myapp",  # lowercase
        "myApp",  # contains uppercase
        "MyApp",  # initial uppercase
        "MyAPP",  # ends in uppercase
        "my-app",  # contains hyphen
        "my_app",  # contains underscore
        "myapp2",  # ends with digit
        "my2app",  # contains digit
    ],
)
def test_valid_app_name(name):
    try:
        AppConfig(
            app_name=name,
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/" + name.replace("-", "_")],
            license={"file": "LICENSE"},
        )
    except BriefcaseConfigError:
        pytest.fail(f"{name} should be valid")


@pytest.mark.parametrize(
    "name",
    [
        "!myapp",  # initial punctuation
        "my!app",  # contains punctuation
        "myapp!",  # end punctuation
        "my$app",  # other punctuation
        "-myApp",  # initial hyphen
        "myApp-",  # end hyphen
        "_myApp",  # initial underscore
        "myApp_",  # end underscore
    ],
)
def test_invalid_app_name(name):
    with pytest.raises(BriefcaseConfigError, match=r"is not a valid app name\."):
        AppConfig(
            app_name=name,
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/invalid"],
            license={"file": "LICENSE"},
        )


@pytest.mark.parametrize(
    "bundle",
    [
        "com.example",
        "com.example.more",
        "com.example42.more",
        "com.example-42.more",
    ],
)
def test_valid_bundle(bundle):
    try:
        AppConfig(
            app_name="myapp",
            version="1.2.3",
            bundle=bundle,
            description="A simple app",
            sources=["src/myapp"],
            license={"file": "LICENSE"},
        )
    except BriefcaseConfigError:
        pytest.fail(f"{bundle} should be valid")


@pytest.mark.parametrize(
    "bundle",
    [
        "not a bundle!",  # Free text.
        "home",  # Only one section.
        "com.hello_world",  # underscore
        "com.hello,world",  # comma
        "com.hello world!",  # exclamation point
    ],
)
def test_invalid_bundle_identifier(bundle):
    with pytest.raises(
        BriefcaseConfigError, match=r"is not a valid bundle identifier\."
    ):
        AppConfig(
            app_name="myapp",
            version="1.2.3",
            bundle=bundle,
            description="A simple app",
            sources=["src/invalid"],
            license={"file": "LICENSE"},
        )


def test_valid_app_version():
    try:
        AppConfig(
            app_name="myapp",
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/myapp"],
            license={"file": "LICENSE"},
        )
    except BriefcaseConfigError:
        pytest.fail("1.2.3 should be a valid version number")


def test_invalid_app_version():
    with pytest.raises(
        BriefcaseConfigError,
        match=r"Version number for 'myapp' \(foobar\) is not valid\.",
    ):
        AppConfig(
            app_name="myapp",
            version="foobar",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/invalid"],
            license={"file": "LICENSE"},
        )


@pytest.mark.parametrize(
    "name, module_name",
    [
        ("myapp", "myapp"),
        ("my-app", "my_app"),
    ],
)
def test_module_name(name, module_name):
    config = AppConfig(
        app_name=name,
        version="1.2.3",
        bundle="org.beeware",
        description="A simple app",
        sources=["src/" + module_name],
        license={"file": "LICENSE"},
    )

    assert config.module_name == module_name


@pytest.mark.parametrize(
    "bundle, package_name",
    [
        ("com.example", "com.example"),
        ("com.ex-ample", "com.ex_ample"),
    ],
)
def test_package_name(bundle, package_name):
    config = AppConfig(
        app_name="myapp",
        version="1.2.3",
        bundle=bundle,
        description="A simple app",
        sources=["src/myapp"],
        license={"file": "LICENSE"},
    )

    assert config.package_name == package_name


@pytest.mark.parametrize(
    "app_name, bundle_name",
    [
        ("my-app", "my-app"),
        ("my_app", "my-app"),
    ],
)
def test_bundle_name(app_name, bundle_name):
    config = AppConfig(
        app_name=app_name,
        version="1.2.3",
        bundle="com.example",
        description="A simple app",
        sources=["src/my_app"],
        license={"file": "LICENSE"},
    )

    assert config.bundle_name == bundle_name


@pytest.mark.parametrize(
    "app_name, bundle_name",
    [
        ("my-app", "my-app"),
        ("my_app", "my-app"),
    ],
)
def test_bundle_identifier(app_name, bundle_name):
    bundle = "com.example"

    config = AppConfig(
        app_name=app_name,
        version="1.2.3",
        bundle=bundle,
        description="A simple app",
        sources=["src/my_app"],
        license={"file": "LICENSE"},
    )

    assert config.bundle_identifier == f"{bundle}.{bundle_name}"


@pytest.mark.parametrize(
    "sources",
    [
        ["src/dupe", "src/dupe"],
        ["src/dupe", "src/other", "src/dupe"],
        ["src/dupe", "somewhere/dupe", "src/other"],
        ["src/dupe", "src/deep/dupe", "src/other"],
    ],
)
def test_duplicated_source(sources):
    with pytest.raises(
        BriefcaseConfigError, match=r"contains duplicated package names\."
    ):
        AppConfig(
            app_name="dupe",
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=sources,
            license={"file": "LICENSE"},
        )


def test_no_source_for_app():
    with pytest.raises(
        BriefcaseConfigError, match=r" does not include a package named 'my_app'\."
    ):
        AppConfig(
            app_name="my-app",
            version="1.2.3",
            bundle="org.beeware",
            description="A simple app",
            sources=["src/something", "src/other"],
            license={"file": "LICENSE"},
        )



import pytest

from briefcase.console import Console, Log
from briefcase.platforms.iOS.xcode import iOSXcodePackageCommand


@pytest.fixture
def package_command(tmp_path, first_app_config):
    command = iOSXcodePackageCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.data_path = tmp_path / "briefcase"

    return command


def test_packaging_formats(package_command):
    assert package_command.packaging_formats == ["ipa"]


def test_default_packaging_format(package_command):
    assert package_command.default_packaging_format == "ipa"



import pytest

from briefcase.console import LogLevel

from ...utils import create_file


@pytest.fixture
def myapp_unrolled(myapp, support_path, app_packages_path_index):
    # Create some files that can be cleaned up
    create_file(support_path / "dir1/a_file1.txt", "pork")
    create_file(support_path / "dir1/a_file2.doc", "ham")
    create_file(support_path / "dir1/b_file.txt", "eggs")
    create_file(support_path / "dir1/__pycache__/first.pyc", "pyc 1")
    create_file(support_path / "dir1/__pycache__/second.pyc", "pyc 2")
    create_file(support_path / "dir2/b_file.txt", "spam")
    create_file(support_path / "other/deep/b_file.doc", "wigs")
    create_file(support_path / "other/deep/other.doc", "wigs")

    return myapp


@pytest.mark.parametrize("debug", [True, False])
def test_no_cleanup(create_command, myapp_unrolled, support_path, debug, capsys):
    """If there are no cleanup directives, bundle content isn't touched; but __pycache__
    is cleaned."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the files are still there, except for the pycache
    assert (support_path / "dir1/a_file1.txt").exists()
    assert (support_path / "dir1/a_file2.doc").exists()
    assert (support_path / "dir1/b_file.txt").exists()
    assert not (support_path / "dir1/__pycache__").exists()
    assert (support_path / "dir2/b_file.txt").exists()
    assert (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (5 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_dir_cleanup(create_command, myapp_unrolled, support_path, debug, capsys):
    """A directory can be cleaned up."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    myapp_unrolled.cleanup_paths = ["path/to/support/dir1"]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm that dir1 has been removed
    assert not (support_path / "dir1").exists()
    assert (support_path / "dir2/b_file.txt").exists()
    assert (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (5 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_file_cleanup(create_command, myapp_unrolled, support_path, debug, capsys):
    """A single file can be cleaned up."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    myapp_unrolled.cleanup_paths = ["path/to/support/dir1/a_file1.txt"]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the named file (plus __pycache__) has been removed
    assert not (support_path / "dir1/a_file1.txt").exists()
    assert (support_path / "dir1/a_file2.doc").exists()
    assert (support_path / "dir1/b_file.txt").exists()
    assert (support_path / "dir2/b_file.txt").exists()
    assert not (support_path / "dir1/__pycache__").exists()
    assert (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (6 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_all_files_in_dir_cleanup(
    create_command, myapp_unrolled, support_path, debug, capsys
):
    """All files in a directory can be cleaned up."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    myapp_unrolled.cleanup_paths = ["path/to/support/dir1/*"]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the named files (and __pycache__) have been removed,
    # but the dir still exists
    assert not (support_path / "dir1/a_file1.txt").exists()
    assert not (support_path / "dir1/a_file2.doc").exists()
    assert not (support_path / "dir1/b_file.txt").exists()
    assert not (support_path / "dir1/__pycache__").exists()
    assert (support_path / "dir1").exists()
    assert (support_path / "dir2/b_file.txt").exists()
    assert (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (8 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_dir_glob_cleanup(create_command, myapp_unrolled, support_path, debug, capsys):
    """A glob of directories can be cleaned up."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    myapp_unrolled.cleanup_paths = ["path/to/support/dir*"]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the matching directories have been removed
    assert not (support_path / "dir1").exists()
    assert not (support_path / "dir2").exists()
    assert (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (6 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_file_glob_cleanup(create_command, myapp_unrolled, support_path, debug, capsys):
    """A glob of files can be cleaned up."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    myapp_unrolled.cleanup_paths = ["path/to/support/dir1/*.txt"]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the matching files (plus __pycache__) have been removed
    assert not (support_path / "dir1/a_file1.txt").exists()
    assert (support_path / "dir1/a_file2.doc").exists()
    assert not (support_path / "dir1/b_file.txt").exists()
    assert not (support_path / "dir1/__pycache__").exists()
    assert (support_path / "dir2/b_file.txt").exists()
    assert (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (7 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_deep_glob_cleanup(create_command, myapp_unrolled, support_path, debug, capsys):
    """A glob that matches all directories will be added to the cleanup list."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    myapp_unrolled.cleanup_paths = ["path/to/support/**/b_file.*"]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the matching files (plus __pycache__) have been removed
    assert (support_path / "dir1/a_file1.txt").exists()
    assert (support_path / "dir1/a_file2.doc").exists()
    assert not (support_path / "dir1/b_file.txt").exists()
    assert not (support_path / "dir1/__pycache__").exists()
    assert not (support_path / "dir2/b_file.txt").exists()
    assert not (support_path / "other/deep/b_file.doc").exists()
    assert (support_path / "other/deep/other.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (8 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_template_glob_cleanup(
    create_command, myapp_unrolled, support_path, debug, capsys
):
    """A glob of files specified in the template will be added to the cleanup list."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    # Define a cleanup_paths in the template *and* on the app
    create_command._briefcase_toml[myapp_unrolled] = {
        "paths": {
            "cleanup_paths": ["path/to/support/dir1/a_*.*"],
        }
    }
    myapp_unrolled.cleanup_paths = ["path/to/support/other/*"]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the files from the app config and template config have been
    # removed, as well as __pycache__
    assert not (support_path / "dir1/a_file1.txt").exists()
    assert not (support_path / "dir1/a_file2.doc").exists()
    assert (support_path / "dir1/b_file.txt").exists()
    assert not (support_path / "dir1/__pycache__").exists()
    assert (support_path / "dir2/b_file.txt").exists()
    assert not (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (8 if debug else 4)


@pytest.mark.parametrize("debug", [True, False])
def test_non_existent_cleanup(
    create_command,
    myapp_unrolled,
    support_path,
    debug,
    capsys,
):
    """Referencing a specific file that doesn't exist doesn't cause a problem."""
    if debug:
        create_command.logger.verbosity = LogLevel.DEBUG

    myapp_unrolled.cleanup_paths = [
        # This file exists
        "path/to/support/dir1/a_file1.txt",
        # None of these paths do.
        "path/to/support/dir1/missing.txt",
        "path/to/support/nowhere",
        "path/to/support/absent/*",
    ]

    # Cleanup app content
    create_command.cleanup_app_content(myapp_unrolled)

    # Confirm the single existing file named has been removed,
    # as well as __pycache__
    assert not (support_path / "dir1/a_file1.txt").exists()
    assert (support_path / "dir1/a_file2.doc").exists()
    assert (support_path / "dir1/b_file.txt").exists()
    assert not (support_path / "dir1/__pycache__").exists()
    assert (support_path / "dir2/b_file.txt").exists()
    assert (support_path / "other/deep/b_file.doc").exists()

    # Console output ends with the done message; the number of other messages depends on
    # whether debug is enabled.
    output = capsys.readouterr().out.split("\n")
    assert output[-3] == "Removing unneeded app bundle content... done"
    assert len(output) == (6 if debug else 4)



import subprocess
from unittest.mock import call

import pytest

from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.usefixtures("mock_docker")
def test_cache_image(mock_tools, sub_kw, sub_check_output_kw):
    """A Docker image can be cached."""
    # mock image not being cached in Docker
    mock_tools.subprocess._subprocess.check_output.return_value = ""

    # Cache an image
    mock_tools.docker.cache_image("ubuntu:jammy")

    # Confirms that image is not available
    mock_tools.subprocess._subprocess.check_output.assert_called_with(
        ["docker", "images", "-q", "ubuntu:jammy"],
        env={"PROCESS_ENV_VAR": "VALUE", "DOCKER_CLI_HINTS": "false"},
        **sub_check_output_kw,
    )

    # Image is pulled and cached
    mock_tools.subprocess._subprocess.run.assert_has_calls(
        [
            call(
                ["docker", "pull", "ubuntu:jammy"],
                check=True,
                env={"PROCESS_ENV_VAR": "VALUE", "DOCKER_CLI_HINTS": "false"},
                **sub_kw,
            ),
        ]
    )


@pytest.mark.usefixtures("mock_docker")
def test_cache_image_already_cached(mock_tools, sub_check_output_kw):
    """A Docker image is not pulled if it is already cached."""
    # mock image already cached in Docker
    mock_tools.subprocess._subprocess.check_output.return_value = "99284ca6cea0"

    # Cache an image
    mock_tools.docker.cache_image("ubuntu:jammy")

    # Confirms that image is not available
    mock_tools.subprocess._subprocess.check_output.assert_called_with(
        ["docker", "images", "-q", "ubuntu:jammy"],
        env={"PROCESS_ENV_VAR": "VALUE", "DOCKER_CLI_HINTS": "false"},
        **sub_check_output_kw,
    )

    # Image is not pulled and cached
    mock_tools.subprocess._subprocess.run.assert_has_calls([])


@pytest.mark.usefixtures("mock_docker")
def test_cache_bad_image(mock_tools):
    """If an image is invalid, an exception raised."""
    # mock image not being cached in Docker
    mock_tools.subprocess._subprocess.check_output.return_value = ""

    # Mock a Docker failure due to a bad image
    mock_tools.subprocess._subprocess.run.side_effect = subprocess.CalledProcessError(
        returncode=125, cmd="docker..."
    )

    # Try to cache an image that doesn't exist
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to obtain the Docker image for ubuntu:does-not-exist.",
    ):
        mock_tools.docker.cache_image("ubuntu:does-not-exist")



from unittest import mock

from briefcase.config import AppConfig


def test_no_resources(create_command):
    """If the template defines no extra targets, none are installed."""
    myapp = AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        license={"file": "LICENSE"},
    )

    # Prime the path index with no targets
    create_command._briefcase_toml[myapp] = {"paths": {}}

    install_image = mock.MagicMock()
    create_command.install_image = install_image

    # Install app resources
    create_command.install_app_resources(myapp)

    # No icons or document types, so no calls to install images
    install_image.assert_not_called()


def test_icon_target(create_command, tmp_path):
    """If the template defines an icon target, it will be installed."""
    myapp = AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        icon="images/icon",
        license={"file": "LICENSE"},
    )

    # Prime the path index with 2 icon targets
    create_command._briefcase_toml[myapp] = {
        "paths": {
            "icon": {
                "10": "path/to/icon-10.png",
                "20": "path/to/icon-20.png",
            }
        }
    }

    install_image = mock.MagicMock()
    create_command.install_image = install_image

    # Install app resources
    create_command.install_app_resources(myapp)

    # 2 calls to install icons will be made
    install_image.assert_has_calls(
        [
            mock.call(
                "application icon",
                source="images/icon",
                variant=None,
                size="10",
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "icon-10.png",
            ),
            mock.call(
                "application icon",
                source="images/icon",
                variant=None,
                size="20",
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "icon-20.png",
            ),
        ],
        any_order=True,
    )


def test_icon_variant_target(create_command, tmp_path):
    """If the template defines an icon target with variants, they will be installed."""
    myapp = AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        icon={
            "round": "images/round",
            "square": "images/square",
        },
        license={"file": "LICENSE"},
    )

    # Prime the path index with 2 icon targets
    create_command._briefcase_toml[myapp] = {
        "paths": {
            "icon": {
                "round": "path/to/round.png",
                "square": {
                    "10": "path/to/square-10.png",
                    "20": "path/to/square-20.png",
                },
            }
        }
    }

    install_image = mock.MagicMock()
    create_command.install_image = install_image

    # Install app resources
    create_command.install_app_resources(myapp)

    # 3 calls to install icons will be made
    install_image.assert_has_calls(
        [
            mock.call(
                "application icon",
                source={"round": "images/round", "square": "images/square"},
                variant=None,
                size="round",  # This is expected for unsized variants
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "round.png",
            ),
            mock.call(
                "application icon",
                source={"round": "images/round", "square": "images/square"},
                variant="square",
                size="10",
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "square-10.png",
            ),
            mock.call(
                "application icon",
                source={"round": "images/round", "square": "images/square"},
                variant="square",
                size="20",
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "square-20.png",
            ),
        ],
        any_order=True,
    )


def test_splash_target(create_command, capsys):
    """If the template defines a splash target, a warning will be raised."""
    myapp = AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        splash="images/splash",
        license={"file": "LICENSE"},
    )

    # Prime an empty path index
    create_command._briefcase_toml[myapp] = {"paths": {}}

    install_image = mock.MagicMock()
    create_command.install_image = install_image

    # Install app resources
    create_command.install_app_resources(myapp)

    # No calls to install splashes are made
    install_image.assert_not_called()

    # A warning about the splash configuration was raised.
    assert "The splash configuration will be ignored." in capsys.readouterr().out


def test_splash_variant_target(create_command, capsys):
    """If the template defines a splash target with variants, a warning is raised."""
    myapp = AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        splash={
            "portrait": "images/portrait",
            "landscape": "images/landscape",
        },
        license={"file": "LICENSE"},
    )

    # Prime an empty path index
    create_command._briefcase_toml[myapp] = {"paths": {}}

    install_image = mock.MagicMock()
    create_command.install_image = install_image

    # Install app resources
    create_command.install_app_resources(myapp)

    # No calls to install splashes are made
    install_image.assert_not_called()

    # A warning about the splash configuration was raised.
    assert "The splash configuration will be ignored." in capsys.readouterr().out


def test_doctype_icon_target(create_command, tmp_path):
    """If the template defines document types, their icons will be installed."""
    myapp = AppConfig(
        app_name="my-app",
        formal_name="My App",
        bundle="com.example",
        version="1.2.3",
        description="This is a simple app",
        sources=["src/my_app"],
        document_type={
            "mydoc": {"icon": "images/mydoc-icon"},
            "other": {
                "icon": "images/other-icon",
            },
        },
        license={"file": "LICENSE"},
    )

    # Prime the path index with 2 document types;
    # * mydoc, which has a single static image; and
    # * other, which has multiple size targets.
    create_command._briefcase_toml[myapp] = {
        "paths": {
            "document_type_icon": {
                "mydoc": "path/to/mydoc-icon.png",
                "other": {
                    "10": "path/to/other-icon-10.png",
                    "20": "path/to/other-icon-20.png",
                },
            }
        }
    }

    install_image = mock.MagicMock()
    create_command.install_image = install_image

    # Install app resources
    create_command.install_app_resources(myapp)

    # 2 calls to install doctype icon images will be made
    install_image.assert_has_calls(
        [
            mock.call(
                "icon for .mydoc documents",
                source="images/mydoc-icon",
                variant=None,
                size=None,
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "mydoc-icon.png",
            ),
            mock.call(
                "icon for .other documents",
                source="images/other-icon",
                variant=None,
                size="10",
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "other-icon-10.png",
            ),
            mock.call(
                "icon for .other documents",
                source="images/other-icon",
                variant=None,
                size="20",
                target=tmp_path
                / "base_path"
                / "build"
                / "my-app"
                / "tester"
                / "dummy"
                / "path"
                / "to"
                / "other-icon-20.png",
            ),
        ],
        any_order=True,
    )



from subprocess import CalledProcessError
from unittest.mock import MagicMock

import pytest

from briefcase.console import LogLevel
from briefcase.exceptions import BriefcaseCommandError

from ....utils import create_file


@pytest.fixture
def first_app_apk(first_app_config):
    first_app_config.packaging_format = "apk"
    return first_app_config


def test_unsupported_template_version(package_command, first_app_generated):
    """Error raised if template's target version is not supported."""
    # Mock the build command previously called
    create_file(package_command.binary_path(first_app_generated), content="")

    package_command.verify_app = MagicMock(wraps=package_command.verify_app)

    package_command._briefcase_toml.update(
        {first_app_generated: {"briefcase": {"target_epoch": "0.3.16"}}}
    )

    with pytest.raises(
        BriefcaseCommandError,
        match="The app template used to generate this app is not compatible",
    ):
        package_command(first_app_generated, packaging_format="apk")

    package_command.verify_app.assert_called_once_with(first_app_generated)


def test_distribution_path(package_command, first_app_apk, tmp_path):
    print(package_command.packaging_formats)
    assert (
        package_command.distribution_path(first_app_apk)
        == tmp_path / "base_path/dist/First App-0.0.1.apk"
    )


@pytest.mark.parametrize(
    "host_os, gradlew_name, tool_debug_mode",
    [
        ("Windows", "gradlew.bat", True),
        ("Windows", "gradlew.bat", False),
        ("NonWindows", "gradlew", True),
        ("NonWindows", "gradlew", False),
    ],
)
def test_execute_gradle(
    package_command,
    first_app_apk,
    host_os,
    gradlew_name,
    tool_debug_mode,
    tmp_path,
):
    """Validate that package_app() will launch `gradlew bundleRelease` with the
    appropriate environment & cwd, and that it will use `gradlew.bat` on Windows but
    `gradlew` elsewhere."""
    # Mock out `host_os` so we can validate which name is used for gradlew.
    package_command.tools.host_os = host_os

    # Enable verbose tool logging
    if tool_debug_mode:
        package_command.tools.logger.verbosity = LogLevel.DEEP_DEBUG

    # Set up a side effect of invoking gradle to create a bundle
    def create_bundle(*args, **kwargs):
        create_file(
            tmp_path
            / "base_path"
            / "build"
            / "first-app"
            / "android"
            / "gradle"
            / "app"
            / "build"
            / "outputs"
            / "apk"
            / "release"
            / "app-release-unsigned.apk",
            "Android release",
        )

    package_command.tools.subprocess.run.side_effect = create_bundle

    # Create mock environment with `key`, which we expect to be preserved, and
    # `ANDROID_SDK_ROOT`, which we expect to be overwritten.
    package_command.tools.os.environ = {"ANDROID_SDK_ROOT": "somewhere", "key": "value"}

    package_command.package_app(first_app_apk)

    package_command.tools.android_sdk.verify_emulator.assert_called_once_with()
    package_command.tools.subprocess.run.assert_called_once_with(
        [
            package_command.bundle_path(first_app_apk) / gradlew_name,
            "--console",
            "plain",
        ]
        + (["--debug"] if tool_debug_mode else [])
        + ["assembleRelease"],
        cwd=package_command.bundle_path(first_app_apk),
        env=package_command.tools.android_sdk.env,
        check=True,
        encoding="ISO-42",
    )

    # The release asset has been moved into the dist folder
    assert (tmp_path / "base_path/dist/First App-0.0.1.apk").exists()


def test_print_gradle_errors(package_command, first_app_apk):
    """Validate that package_app() will convert stderr/stdout from the process into
    exception text."""
    # Create a mock subprocess that crashes, printing text partly in non-ASCII.
    package_command.tools.subprocess.run.side_effect = CalledProcessError(
        returncode=1,
        cmd=["ignored"],
    )
    with pytest.raises(BriefcaseCommandError):
        package_command.package_app(first_app_apk)



from __future__ import annotations

import sys
from operator import attrgetter

from briefcase.exceptions import (
    BriefcaseCommandError,
    UnsupportedHostError,
    UpgradeToolError,
)
from briefcase.integrations.base import ManagedTool, Tool, tool_registry

from .base import BaseCommand


class UpgradeCommand(BaseCommand):
    cmd_line = "briefcase upgrade"
    command = "upgrade"
    output_format = None
    description = "Upgrade Briefcase-managed tools."

    @property
    def platform(self):
        """The upgrade command always reports as the local platform."""
        return {
            "darwin": "macOS",
            "linux": "linux",
            "win32": "windows",
        }[sys.platform]

    def bundle_path(self, app):
        """A placeholder; Upgrade command doesn't have a bundle path."""
        raise NotImplementedError()

    def binary_path(self, app):
        """A placeholder; Upgrade command doesn't have a binary path."""
        raise NotImplementedError()

    def add_options(self, parser):
        parser.add_argument(
            "-l",
            "--list",
            dest="list_tools",
            action="store_true",
            help="List the Briefcase-managed tools that are currently installed",
        )
        parser.add_argument(
            "tool_list",
            metavar="tool",
            nargs="*",
            help="The Briefcase-managed tool to upgrade. If no tool is named, all tools will be upgraded",
        )

    def get_tools_to_upgrade(self, tool_list: set[str]) -> list[ManagedTool]:
        """Returns set of managed Tools that can be upgraded.

        Raises ``BriefcaseCommandError`` if user list contains any invalid tool names.
        """
        upgrade_list: set[type[Tool]]
        tools_to_upgrade: set[ManagedTool] = set()

        # Validate user tool list against tool registry
        if tool_list:
            if invalid_tools := tool_list - set(tool_registry):
                raise UpgradeToolError(
                    f"Briefcase does not know how to manage {', '.join(sorted(invalid_tools))}."
                )
            upgrade_list = {
                tool for name, tool in tool_registry.items() if name in tool_list
            }
        else:
            upgrade_list = set(tool_registry.values())

        # Filter list of tools to those that are being managed
        for tool_klass in upgrade_list:
            if issubclass(tool_klass, ManagedTool):
                try:
                    tool = tool_klass.verify(tools=self.tools, install=False)
                except (BriefcaseCommandError, UnsupportedHostError):
                    pass
                else:
                    if tool.managed_install:
                        tools_to_upgrade.add(tool)

        # Let the user know if any requested tools are not being managed
        if tool_list:
            if unmanaged_tools := tool_list - {tool.name for tool in tools_to_upgrade}:
                error_msg = (
                    f"Briefcase is not managing {', '.join(sorted(unmanaged_tools))}."
                )
                if not tools_to_upgrade:
                    raise UpgradeToolError(error_msg)
                else:
                    self.logger.warning(error_msg)

        return sorted(list(tools_to_upgrade), key=attrgetter("name"))

    def __call__(self, tool_list: list[str], list_tools: bool = False, **options):
        """Perform tool upgrades or list tools qualifying for upgrade.

        :param tool_list: List of tool names from user to upgrade.
        :param list_tools: Boolean to only list upgradeable tools (default False).
        """
        if tools_to_upgrade := self.get_tools_to_upgrade(set(tool_list)):
            self.logger.info(
                f"Briefcase {'is managing' if list_tools else 'will upgrade'} the following tools:",
                prefix=self.command,
            )
            for tool in tools_to_upgrade:
                self.logger.info(f" - {tool.full_name} ({tool.name})")

            if not list_tools:
                for tool in tools_to_upgrade:
                    self.logger.info(f"Upgrading {tool.full_name}...", prefix=tool.name)
                    tool.upgrade()
        else:
            self.logger.info("Briefcase is not managing any tools.")



import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_single_source(base_command, my_app):
    """If an app provides a single source location and it matches, it is selected as the
    dist-info location."""
    my_app.sources = ["src/my_app"]

    assert base_command.app_module_path(my_app) == base_command.base_path / "src/my_app"


def test_no_prefix(base_command, my_app):
    """If an app provides a source location without a prefix and it matches, it is
    selected as the dist-info location."""
    my_app.sources = ["my_app"]

    assert base_command.app_module_path(my_app) == base_command.base_path / "my_app"


def test_long_prefix(base_command, my_app):
    """If an app provides a source location with a long prefix and it matches, it is
    selected as the dist-info location."""
    my_app.sources = ["path/to/src/my_app"]

    assert (
        base_command.app_module_path(my_app)
        == base_command.base_path / "path/to/src/my_app"
    )


def test_matching_source(base_command, my_app):
    """If an app provides a single matching source location, it is selected as the dist-
    info location."""
    my_app.sources = ["src/other", "src/my_app", "src/extra"]

    assert base_command.app_module_path(my_app) == base_command.base_path / "src/my_app"


def test_multiple_match(base_command, my_app):
    """If an app provides multiple matching source location, an error is raised."""
    my_app.sources = ["src/my_app", "extra/my_app"]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Multiple paths in sources found for application 'my-app'",
    ):
        base_command.app_module_path(my_app)


def test_hyphen_source(base_command, my_app):
    """If an app provides a single source location with a hyphen, an error is raised."""
    # The source directory must be a valid module, so hyphens aren't legal.
    my_app.sources = ["src/my-app"]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to find code for application 'my-app'",
    ):
        base_command.app_module_path(my_app)


def test_no_match(base_command, my_app):
    """If an app provides a multiple locations, none of which match, an error is
    raised."""
    my_app.sources = ["src/pork", "src/spam"]

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to find code for application 'my-app'",
    ):
        base_command.app_module_path(my_app)


def test_no_source(base_command, my_app):
    """If an app provides no source locations, an error is raised."""
    my_app.sources = []

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Unable to find code for application 'my-app'",
    ):
        base_command.app_module_path(my_app)



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.cookiecutter import PythonVersionExtension


@pytest.mark.parametrize(
    "value, expected",
    [
        # Single digit minor
        ("3.8.4.dev5", "3.8"),
        ("3.8.4a1", "3.8"),
        ("3.8.4b2", "3.8"),
        ("3.8.4rc3", "3.8"),
        ("3.8.4.post6", "3.8"),
        # Two digit minor
        ("3.11.4.dev5", "3.11"),
        ("3.11.4a1", "3.11"),
        ("3.11.4b2", "3.11"),
        ("3.11.4rc3", "3.11"),
        ("3.11.4.post6", "3.11"),
    ],
)
def test_py_tag(value, expected):
    env = MagicMock()
    env.filters = {}
    PythonVersionExtension(env)
    assert env.filters["py_tag"](value) == expected


@pytest.mark.parametrize(
    "value, expected",
    [
        # Single digit minor
        ("3.8.4.dev5", "38"),
        ("3.8.4a1", "38"),
        ("3.8.4b2", "38"),
        ("3.8.4rc3", "38"),
        ("3.8.4.post6", "38"),
        # Two digit minor
        ("3.11.4.dev5", "311"),
        ("3.11.4a1", "311"),
        ("3.11.4b2", "311"),
        ("3.11.4rc3", "311"),
        ("3.11.4.post6", "311"),
    ],
)
def test_py_libtag(value, expected):
    env = MagicMock()
    env.filters = {}
    PythonVersionExtension(env)
    assert env.filters["py_libtag"](value) == expected



import os
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.console import Console, Log
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.windows.visualstudio import WindowsVisualStudioOpenCommand

from ....utils import create_file


@pytest.fixture
def open_command(tmp_path, first_app_config):
    command = WindowsVisualStudioOpenCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.os = MagicMock(spec_set=os)
    command.tools.subprocess = MagicMock(spec_set=Subprocess)
    return command


@pytest.mark.skipif(sys.platform != "win32", reason="Windows specific test")
def test_open(open_command, first_app_config, tmp_path):
    """Open command opens visual studio project file."""
    # Create the solution file that would be in the created project.
    create_file(open_command.project_path(first_app_config), "Visual Studio Solution")

    open_command(first_app_config)

    open_command.tools.os.startfile.assert_called_once_with(
        tmp_path
        / "base_path"
        / "build"
        / "first-app"
        / "windows"
        / "visualstudio"
        / "First App.sln"
    )



import pytest

from briefcase.commands.create import Git
from briefcase.exceptions import BriefcaseCommandError


def test_no_git(update_command, monkeypatch):
    """If Git is not installed, an error is raised."""

    def monkeypatch_verify_git(*a, **kw):
        raise BriefcaseCommandError("Briefcase requires git, but it is not installed")

    monkeypatch.setattr(Git, "verify", monkeypatch_verify_git)

    # The command will fail tool verification.
    with pytest.raises(
        BriefcaseCommandError, match=r"Briefcase requires git, but it is not installed"
    ):
        update_command()


def test_update(update_command, first_app, second_app):
    """The update command can be called."""
    # Configure no command line options
    options, _ = update_command.parse_options([])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("cleanup", "second"),
    ]


def test_update_single(update_command, first_app, second_app):
    """The update command can be called to update a single app from the config."""
    # Configure no command line options
    options, _ = update_command.parse_options([])

    update_command(app=update_command.apps["first"], **options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup", "first"),
    ]


def test_update_with_requirements(update_command, first_app, second_app):
    """The update command can be called, requesting a requirements update."""
    # Configure a requirements update
    options, _ = update_command.parse_options(["-r"])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("requirements", "first", False),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("requirements", "second", False),
        ("cleanup", "second"),
    ]


def test_update_with_resources(update_command, first_app, second_app):
    """The update command can be called, requesting a resources update."""
    # Configure no command line options
    options, _ = update_command.parse_options(["--update-resources"])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("resources", "first"),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("resources", "second"),
        ("cleanup", "second"),
    ]


def test_update_with_support(update_command, first_app, second_app):
    """The update command can be called, requesting an app support update."""
    # Configure no command line options
    options, _ = update_command.parse_options(["--update-support"])

    update_command(**options)

    # The right sequence of things will be done
    assert update_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update the first app
        ("verify-app-template", "first"),
        ("verify-app-tools", "first"),
        ("code", "first", False),
        ("cleanup-support", "first"),
        ("support", "first"),
        ("cleanup", "first"),
        # Update the second app
        ("verify-app-template", "second"),
        ("verify-app-tools", "second"),
        ("code", "second", False),
        ("cleanup-support", "second"),
        ("support", "second"),
        ("cleanup", "second"),
    ]



import pytest

from briefcase.commands.run import LogFilter


@pytest.mark.parametrize(
    "recent_history, returncode",
    (
        # Zero return code
        [">>>>>>>>>> EXIT 0 <<<<<<<<<<", 0],
        # positive integer return code
        [">>>>>>>>>> EXIT 123 <<<<<<<<<<", 123],
        # Zero return code
        [">>>>>>>>>> EXIT -15 <<<<<<<<<<", -15],
        # Non-alpha return code
        [">>>>>>>>>> EXIT abc <<<<<<<<<<", -999],
        # Mixed return code
        [">>>>>>>>>> EXIT 42 and some more <<<<<<<<<<", -999],
    ),
)
def test_default_exit_filter(recent_history, returncode):
    """The default exit filter captures exit criteria."""
    exit_func = LogFilter.test_filter(LogFilter.DEFAULT_EXIT_REGEX)

    tail = "\n".join(["line 1", "line 2"] + [recent_history])
    assert exit_func(tail) == returncode


@pytest.mark.parametrize(
    "recent_history",
    (
        # No exit
        "This doesn't match",
        # Not enough chevrons
        ">>>> EXIT 3 <<<<",
        # Wrong text
        ">>>>>>>>>> EXEUNT 3 <<<<<<<<<<",
        # Extra text
        ">>>>>>>>>> EXIT 3 <<<<<<<<<< but there's more!",
    ),
)
def test_default_exit_filter_no_match(recent_history):
    """The default exit filter *doesn't* catch content that doesn't match the regex."""
    exit_func = LogFilter.test_filter(LogFilter.DEFAULT_EXIT_REGEX)

    tail = "\n".join(["line 1", "line 2"] + [recent_history])
    assert exit_func(tail) is None


def test_custom_filter():
    """The user can specify a custom exit filter."""
    custom_func = LogFilter.test_filter("WIBBLE (?P<returncode>.*) WIBBLE")

    # Custom filter doesn't match "normal" output
    recent = ["line 1", "line 2", ">>>>>>>>>> EXIT 123 <<<<<<<<<<"]
    assert custom_func("\n".join(recent)) is None

    # Custom filter does match "custom" output
    recent = ["line 1", "line 2", "WIBBLE 123 WIBBLE"]
    assert custom_func("\n".join(recent)) == 123


def test_bad_custom_filter():
    """If the custom filter doesn't capture a returncode named group, any match has a
    known return value."""
    custom_func = LogFilter.test_filter(r"WIBBLE \d+ WIBBLE")

    # Custom filter matches, but doesn't capture output
    recent = ["line 1", "line 2", "WIBBLE 123 WIBBLE"]
    assert custom_func("\n".join(recent)) == -998



import subprocess
from unittest.mock import ANY

import pytest

from briefcase.console import LogLevel

# These tests ignore the elsewhere-tested complexities of Dockerizing
# the arguments and just focuses on the semantics of a check_output() call.


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_simple_call(mock_tools, my_app, tmp_path, sub_check_output_kw, capsys):
    """A simple call will be invoked."""
    output = mock_tools[my_app].app_context.check_output(["hello", "world"])

    assert output == "goodbye\n"
    mock_tools[my_app].app_context._dockerize_args.assert_called_once_with(
        ["hello", "world"]
    )
    mock_tools.subprocess._subprocess.check_output.assert_called_with(
        ANY,
        env={"DOCKER_CLI_HINTS": "false", "PROCESS_ENV_VAR": "VALUE"},
        **sub_check_output_kw,
    )
    assert capsys.readouterr().out == ""


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_call_with_extra_kwargs(mock_tools, my_app, tmp_path, capsys):
    """Extra keyword arguments are passed through to subprocess."""
    output = mock_tools[my_app].app_context.check_output(
        ["hello", "world"],
        encoding="ISO-42",
        extra="extra",
    )

    assert output == "goodbye\n"
    mock_tools[my_app].app_context._dockerize_args.assert_called_once_with(
        ["hello", "world"],
        encoding="ISO-42",
        extra="extra",
    )
    mock_tools.subprocess._subprocess.check_output.assert_called_once_with(
        ANY,
        extra="extra",
        encoding="ISO-42",
        stderr=subprocess.STDOUT,
        text=True,
        errors="backslashreplace",
        env={"DOCKER_CLI_HINTS": "false", "PROCESS_ENV_VAR": "VALUE"},
    )
    assert capsys.readouterr().out == ""


@pytest.mark.usefixtures("mock_docker")
@pytest.mark.usefixtures("mock_docker_app_context")
def test_simple_verbose_call(
    mock_tools,
    my_app,
    tmp_path,
    sub_check_output_kw,
    capsys,
):
    """If verbosity is turned out, there is output."""
    mock_tools[my_app].app_context.tools.logger.verbosity = LogLevel.DEBUG

    output = mock_tools[my_app].app_context.check_output(["hello", "world"])

    assert output == "goodbye\n"
    mock_tools[my_app].app_context._dockerize_args.assert_called_once_with(
        ["hello", "world"]
    )
    mock_tools.subprocess._subprocess.check_output.assert_called_once_with(
        ANY,
        env={"DOCKER_CLI_HINTS": "false", "PROCESS_ENV_VAR": "VALUE"},
        **sub_check_output_kw,
    )
    assert ">>> Running Command:\n" in capsys.readouterr().out



from unittest import mock

import pytest

from briefcase.console import Console, Log, LogLevel
from briefcase.integrations.flatpak import Flatpak
from briefcase.integrations.subprocess import Subprocess
from briefcase.platforms.linux.flatpak import LinuxFlatpakRunCommand


@pytest.fixture
def run_command(tmp_path):
    command = LinuxFlatpakRunCommand(
        logger=Log(),
        console=Console(),
        base_path=tmp_path / "base_path",
        data_path=tmp_path / "briefcase",
    )
    command.tools.flatpak = mock.MagicMock(spec_set=Flatpak)
    command.tools.subprocess = mock.MagicMock(spec_set=Subprocess)

    command._stream_app_logs = mock.MagicMock()

    return command


def test_run_gui_app(run_command, first_app_config):
    """A GUI flatpak can be executed."""
    # Set up the log streamer to return a known stream and a good return code
    log_popen = mock.MagicMock()
    run_command.tools.flatpak.run.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # App is executed
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=[],
        stream_output=True,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_output=False,
    )


def test_run_gui_app_with_passthrough(run_command, first_app_config):
    """A GUI flatpak can be executed in debug mode with args."""
    run_command.logger.verbosity = LogLevel.DEBUG

    # Set up the log streamer to return a known stream and a good return code
    log_popen = mock.MagicMock()
    run_command.tools.flatpak.run.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=False,
        passthrough=["foo", "--bar"],
    )

    # App is executed with args
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=["foo", "--bar"],
        stream_output=True,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=False,
        clean_output=False,
    )


def test_run_gui_app_failed(run_command, first_app_config, tmp_path):
    """If there's a problem starting the GUI app, an exception is raised."""
    run_command.tools.flatpak.run.side_effect = OSError

    with pytest.raises(OSError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # The run command was still invoked
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=[],
        stream_output=True,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app(run_command, first_app_config):
    """A console flatpak can be executed."""
    first_app_config.console_app = True

    # Run the app
    run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # App is executed
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=[],
        stream_output=False,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_with_passthrough(run_command, first_app_config):
    """A console flatpak can be executed in debug mode with args."""
    run_command.logger.verbosity = LogLevel.DEBUG
    first_app_config.console_app = True

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=False,
        passthrough=["foo", "--bar"],
    )

    # App is executed with args
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=["foo", "--bar"],
        stream_output=False,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


def test_run_console_app_failed(run_command, first_app_config, tmp_path):
    """If there's a problem starting the console app, an exception is raised."""
    first_app_config.console_app = True

    run_command.tools.flatpak.run.side_effect = OSError

    with pytest.raises(OSError):
        run_command.run_app(first_app_config, test_mode=False, passthrough=[])

    # The run command was still invoked
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=[],
        stream_output=False,
    )

    # No attempt to stream was made
    run_command._stream_app_logs.assert_not_called()


@pytest.mark.parametrize("is_console_app", [True, False])
def test_run_test_mode(run_command, first_app_config, is_console_app):
    """A flatpak can be executed in test mode."""
    # Test mode apps are always streamed
    first_app_config.console_app = is_console_app

    # Set up the log streamer to return a known stream and a good return code
    log_popen = mock.MagicMock()
    run_command.tools.flatpak.run.return_value = log_popen

    # Run the app
    run_command.run_app(first_app_config, test_mode=True, passthrough=[])

    # App is executed
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=[],
        main_module="tests.first_app",
        stream_output=True,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_output=False,
    )


@pytest.mark.parametrize("is_console_app", [True, False])
def test_run_test_mode_with_args(run_command, first_app_config, is_console_app):
    """A flatpak can be executed in test mode with args."""
    # Test mode apps are always streamed
    first_app_config.console_app = is_console_app

    # Set up the log streamer to return a known stream and a good return code
    log_popen = mock.MagicMock()
    run_command.tools.flatpak.run.return_value = log_popen

    # Run the app with args
    run_command.run_app(
        first_app_config,
        test_mode=True,
        passthrough=["foo", "--bar"],
    )

    # App is executed
    run_command.tools.flatpak.run.assert_called_once_with(
        bundle_identifier="com.example.first-app",
        args=["foo", "--bar"],
        main_module="tests.first_app",
        stream_output=True,
    )

    # The streamer was started
    run_command._stream_app_logs.assert_called_once_with(
        first_app_config,
        popen=log_popen,
        test_mode=True,
        clean_output=False,
    )
